"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
from functools import partial
from typing import TYPE_CHECKING, Literal

import numpy as np
from ase.calculators.calculator import Calculator
from ase.stress import full_3x3_to_voigt_6_stress

from fairchem.core.calculate import pretrained_mlip
from fairchem.core.datasets import data_list_collater
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.units.mlip_unit.api.inference import (
    CHARGE_RANGE,
    DEFAULT_CHARGE,
    DEFAULT_SPIN,
    DEFAULT_SPIN_OMOL,
    SPIN_RANGE,
    InferenceSettings,
    UMATask,
)

if TYPE_CHECKING:
    from ase import Atoms

    from fairchem.core.units.mlip_unit import MLIPPredictUnit


class FAIRChemCalculator(Calculator):
    def __init__(
        self,
        predict_unit: MLIPPredictUnit,
        task_name: UMATask | str | None = None,
        seed: int | None = None,  # deprecated
    ):
        """
        Initialize the FAIRChemCalculator from a model MLIPPredictUnit

        Args:
            predict_unit (MLIPPredictUnit): A pretrained MLIPPredictUnit.
            task_name (UMATask or str, optional): Name of the task to use if using a UMA checkpoint.
                Determines default key names for energy, forces, and stress.
                Can be one of 'omol', 'omat', 'oc20', 'odac', or 'omc'.
        Notes:
            - For models that require total charge and spin multiplicity (currently UMA models on omol mode), `charge`
              and `spin` (corresponding to `spin_multiplicity`) are pulled from `atoms.info` during calculations.
                - `charge` must be an integer representing the total charge on the system and can range from -100 to 100.
                - `spin` must be an integer representing the spin multiplicity and can range from 0 to 100.
                - If `task_name="omol"`, and `charge` or `spin` are not set in `atoms.info`, they will default to
                charge=`0` and spin=`1`.
        """

        super().__init__()

        if seed is not None:
            logging.warning(
                "The 'seed' argument is deprecated and will be removed in future versions. "
                "Please set the seed in the MLIPPredictUnit configuration instead."
            )

        if isinstance(task_name, UMATask):
            task_name = task_name.value

        valid_datasets = list(predict_unit.dataset_to_tasks.keys())
        if task_name is not None:
            assert (
                task_name in valid_datasets
            ), f"Given: {task_name}, Valid options are {valid_datasets}"
            self._task = UMATask(task_name)
        elif len(valid_datasets) == 1:
            self._task = UMATask(valid_datasets[0])
        else:
            raise RuntimeError(
                f"A task name must be provided. Valid options are {valid_datasets}"
            )

        self.implemented_properties = [
            task.property for task in predict_unit.dataset_to_tasks[self.task_name]
        ]
        if "energy" in self.implemented_properties:
            self.implemented_properties.append(
                "free_energy"
            )  # Free energy is a copy of energy, see docstring above

        self.predictor = predict_unit

        self.a2g = partial(
            AtomicData.from_ase,
            task_name=self.task_name,
            r_edges=False,
            r_data_keys=["spin", "charge"],
        )

    @property
    def task_name(self) -> str:
        return self._task.value

    @classmethod
    def from_model_checkpoint(
        cls,
        name_or_path: str,
        task_name: UMATask | None = None,
        inference_settings: InferenceSettings | str = "default",
        overrides: dict | None = None,
        device: Literal["cuda", "cpu"] | None = None,
        seed: int = 41,
    ) -> FAIRChemCalculator:
        """Instantiate a FAIRChemCalculator from a checkpoint file.

        Args:
            cls: The class reference
            name_or_path: A model name from fairchem.core.pretrained.available_models or a path to the checkpoint
                file
            task_name: Task name
            inference_settings: Settings for inference. Can be "default" (general purpose) or "turbo"
                (optimized for speed but requires fixed atomic composition). Advanced use cases can
                use a custom InferenceSettings object.
            overrides: Optional dictionary of settings to override default inference settings.
            device: Optional torch device to load the model onto.
            seed: Random seed for reproducibility.
        """

        if name_or_path in pretrained_mlip.available_models:
            predict_unit = pretrained_mlip.get_predict_unit(
                name_or_path,
                inference_settings=inference_settings,
                overrides=overrides,
                device=device,
            )
        elif os.path.isfile(name_or_path):
            predict_unit = pretrained_mlip.load_predict_unit(
                name_or_path,
                inference_settings=inference_settings,
                overrides=overrides,
                device=device,
            )
        else:
            raise ValueError(
                f"{name_or_path=} is not a valid model name or checkpoint path"
            )
        return cls(predict_unit=predict_unit, task_name=task_name, seed=seed)

    def check_state(self, atoms: Atoms, tol: float = 1e-15) -> list:
        """
        Check for any system changes since the last calculation.

        Args:
            atoms (ase.Atoms): The atomic structure to check.
            tol (float): Tolerance for detecting changes.

        Returns:
            list: A list of changes detected in the system.
        """
        state = super().check_state(atoms, tol=tol)
        if (not state) and (self.atoms.info != atoms.info):
            state.append("info")
        return state

    def calculate(
        self, atoms: Atoms, properties: list[str], system_changes: list[str]
    ) -> None:
        """
        Perform the calculation for the given atomic structure.

        Args:
            atoms (Atoms): The atomic structure to calculate properties for.
            properties (list[str]): The list of properties to calculate.
            system_changes (list[str]): The list of changes in the system.

        Notes:
            - `charge` must be an integer representing the total charge on the system and can range from -100 to 100.
            - `spin` must be an integer representing the spin multiplicity and can range from 0 to 100.
            - If `task_name="omol"`, and `charge` or `spin` are not set in `atoms.info`, they will default to `0`.
            - `charge` and `spin` are currently only used for the `omol` head.
            - The `free_energy` is simply a copy of the `energy` and is not the actual electronic free energy.
              It is only set for ASE routines/optimizers that are hard-coded to use this rather than the `energy` key.
        """

        # Our calculators won't work if natoms=0
        if len(atoms) == 0:
            raise ValueError("Atoms object has no atoms inside.")

        # Check if the atoms object has periodic boundary conditions (PBC) set correctly
        self._check_atoms_pbc(atoms)

        # Validate that charge/spin are set correctly for omol, or default to 0 otherwise
        self._validate_charge_and_spin(atoms)

        # Standard call to check system_changes etc
        Calculator.calculate(self, atoms, properties, system_changes)

        if len(atoms) == 1 and sum(atoms.pbc) == 0:
            self.results = self._get_single_atom_energies(atoms)
        else:
            # Convert using the current a2g object
            data_object = self.a2g(atoms)

            # Batch and predict
            batch = data_list_collater([data_object], otf_graph=True)
            pred = self.predictor.predict(batch)

            # Collect the results into self.results
            self.results = {}
            for calc_key in self.implemented_properties:
                if calc_key == "energy":
                    energy = float(pred[calc_key].detach().cpu().numpy()[0])

                    self.results["energy"] = self.results["free_energy"] = (
                        energy  # Free energy is a copy of energy
                    )
                if calc_key == "forces":
                    forces = pred[calc_key].detach().cpu().numpy()
                    self.results["forces"] = forces
                if calc_key == "stress":
                    stress = pred[calc_key].detach().cpu().numpy().reshape(3, 3)
                    stress_voigt = full_3x3_to_voigt_6_stress(stress)
                    self.results["stress"] = stress_voigt

    def _get_single_atom_energies(self, atoms) -> dict:
        """
        Populate output with single atom energies
        """
        if self.predictor.atom_refs is None:
            raise ValueError(
                "Single atom system but no atomic references present. "
                "Please call fairchem.core.pretrained_mlip.get_predict_unit() "
                "with an appropriate checkpoint name."
            )
        logging.warning(
            "Single atom systems are not handled by the model; "
            "the precomputed DFT result is returned. "
            "Spin multiplicity is ignored for monoatomic systems."
        )
        elt = atoms.get_atomic_numbers()[0]
        results = {}

        atom_refs = self.predictor.atom_refs[self.task_name]
        try:
            energy = atom_refs.get(int(elt), {}).get(atoms.info["charge"])
        except AttributeError:
            energy = atom_refs[int(elt)]
        if energy is None:
            raise ValueError("This model has not stored this element with this charge.")
        results["energy"] = energy
        results["forces"] = np.array([[0.0] * 3])
        results["stress"] = np.array([0.0] * 6)
        return results

    def _check_atoms_pbc(self, atoms) -> None:
        """
        Check for invalid PBC conditions

        Args:
            atoms (ase.Atoms): The atomic structure to check.
        """
        if np.all(atoms.pbc) and np.allclose(atoms.cell, 0):
            raise AllZeroUnitCellError
        if np.any(atoms.pbc) and not np.all(atoms.pbc):
            raise MixedPBCError

    def _validate_charge_and_spin(self, atoms: Atoms) -> None:
        """
        Validate and set default values for charge and spin.

        Args:
            atoms (Atoms): The atomic structure containing charge and spin information.
        """

        if "charge" not in atoms.info:
            if self.task_name == UMATask.OMOL.value:
                logging.warning(
                    "task_name='omol' detected, but charge is not set in atoms.info. Defaulting to charge=0. "
                    "Ensure charge is an integer representing the total charge on the system and is within the range -100 to 100."
                )
            atoms.info["charge"] = DEFAULT_CHARGE

        if "spin" not in atoms.info:
            if self.task_name == UMATask.OMOL.value:
                atoms.info["spin"] = DEFAULT_SPIN_OMOL
                logging.warning(
                    "task_name='omol' detected, but spin multiplicity is not set in atoms.info. Defaulting to spin=1. "
                    "Ensure spin is an integer representing the spin multiplicity from 0 to 100."
                )
            else:
                atoms.info["spin"] = DEFAULT_SPIN

        # Validate charge
        charge = atoms.info["charge"]
        if not isinstance(charge, int):
            raise TypeError(
                f"Invalid type for charge: {type(charge)}. Charge must be an integer representing the total charge on the system."
            )
        if not (CHARGE_RANGE[0] <= charge <= CHARGE_RANGE[1]):
            raise ValueError(
                f"Invalid value for charge: {charge}. Charge must be within the range {CHARGE_RANGE[0]} to {CHARGE_RANGE[1]}."
            )

        # Validate spin
        spin = atoms.info["spin"]
        if not isinstance(spin, int):
            raise TypeError(
                f"Invalid type for spin: {type(spin)}. Spin must be an integer representing the spin multiplicity."
            )
        if not (SPIN_RANGE[0] <= spin <= SPIN_RANGE[1]):
            raise ValueError(
                f"Invalid value for spin: {spin}. Spin must be within the range {SPIN_RANGE[0]} to {SPIN_RANGE[1]}."
            )


class MixedPBCError(ValueError):
    """Specific exception example."""

    def __init__(
        self,
        message="Attempted to guess PBC for an atoms object, but the atoms object has PBC set to True for some"
        "dimensions but not others. Please ensure that the atoms object has PBC set to True for all dimensions.",
    ):
        self.message = message
        super().__init__(self.message)


class AllZeroUnitCellError(ValueError):
    """Specific exception example."""

    def __init__(
        self,
        message="Atoms object claims to have PBC set, but the unit cell is identically 0. Please ensure that the atoms"
        "object has a non-zero unit cell.",
    ):
        self.message = message
        super().__init__(self.message)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

import ase.io
import numpy as np
from ase import units
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.verlet import VelocityVerlet
from ase.optimize.lbfgs import LBFGS
from tqdm import tqdm

from fairchem.core.components.calculate import CalculateRunner

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ase.calculators.calculator import Calculator


from ase.md import MDLogger


def get_thermo(filename):
    """
    read thermo logs.
    """
    with open(filename) as f:
        thermo = f.read().splitlines()
        sim_time, Et = [], []
        for i in range(1, len(thermo)):
            t, Etot, _, _, _ = (float(x) for x in thermo[i].split(" ") if x)
            sim_time.append(t)
            Et.append(Etot)
    return np.array(sim_time), np.array(Et)


MD22_TEMP = {
    "AT-AT-CG-CG": 500.0,
    "AT-AT": 500.0,
    "Ac-Ala3-NHMe": 500.0,
    "DHA": 500.0,
    "buckyball-catcher": 400.0,
    "double-walled_nanotube": 400.0,
    "stachyose": 500.0,
}

TM23_TEMP = {
    "Ag": 1235 * 1.25,
    "Au": 1337 * 1.25,
    "Cd": 594 * 1.25,
    "Co": 1768 * 1.25,
    "Cr": 2180 * 1.25,
    "Cu": 1358 * 1.25,
    "Fe": 1811 * 1.25,
    "Hf": 2506 * 1.25,
    "Hg": 234 * 1.25,
    "Ir": 2739 * 1.25,
    "Mn": 1519 * 1.25,
    "Mo": 2896 * 1.25,
    "Nb": 2750 * 1.25,
    "Ni": 1728 * 1.25,
    "Os": 3306 * 1.25,
    "Pd": 1828 * 1.25,
    "Pt": 2041 * 1.25,
    "Re": 3459 * 1.25,
    "Rh": 2237 * 1.25,
    "Ru": 2607 * 1.25,
    "Ta": 3290 * 1.25,
    "Tc": 2430 * 1.25,
    "Ti": 1941 * 1.25,
    "V": 2183 * 1.25,
    "W": 3695 * 1.25,
    "Zn": 693 * 1.25,
    "Zr": 2128 * 1.25,
}


def get_nve_md_data(dataset_root, dataset_name):
    MD22_MOLS = sorted(MD22_TEMP.keys())
    TM23_METALS = sorted(TM23_TEMP.keys())
    if dataset_name == "tm23":
        dataset = [
            (
                ase.io.read(f"{dataset_root}/tm23/{metal}_melt_nequip_test.xyz"),
                TM23_TEMP[metal],
            )
            for metal in TM23_METALS
        ]
    elif dataset_name == "md22":
        dataset = [
            (ase.io.read(f"{dataset_root}/md22/md22_{mol}.xyz"), MD22_TEMP[mol])
            for mol in MD22_MOLS
        ]
    else:
        raise ValueError(f"Unknown dataset name: {dataset_name}")
    return dataset


class NVEMDRunner(CalculateRunner):
    """Perform a single point calculation of several structures/molecules.

    This class handles the single point calculation of atomic structures using a specified calculator,
    processes the input data in chunks, and saves the results.
    """

    result_glob_pattern: ClassVar[str] = "thermo_*-*.log"

    def __init__(
        self,
        calculator: Calculator,
        input_data: Sequence,
        time_step: float,
        steps: float,
        save_frequency: int = 10,
    ):
        """Initialize the SinglePointRunner.

        Args:
            calculator: ASE calculator to use for energy and force calculations
            input_data: Dataset containing atomic structures to process
            calculate_properties: Sequence of properties to calculate
            normalize_properties_by (dict[str, str] | None): Dictionary mapping property names to natoms or a key in
                atoms.info to normalize by
            save_target_properties (Sequence[str] | None): Sequence of target property names to save in the results file
                These properties need to be available using atoms.get_properties or present in the atoms.info dictionary
        """
        self.time_step = time_step
        self.steps = steps
        self.save_frequency = save_frequency
        super().__init__(calculator=calculator, input_data=input_data)

    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        """Perform singlepoint calculations on a subset of structures.

        Splits the input data into chunks and processes the chunk corresponding to job_num.

        Args:
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.

        Returns:
            list[dict[str, Any]] - List of dictionaries containing calculation results
        """
        assert num_jobs == len(
            self.input_data
        ), "num_jobs must be equal to the number of input data"
        atoms, temp = self.input_data[job_num]
        atoms.calc = self.calculator

        # relax
        opt = LBFGS(
            atoms,
            logfile=str(
                Path(self.job_config.metadata.results_dir)
                / f"relax_{num_jobs}-{job_num}.log"
            ),
        )
        opt.run(fmax=0.05, steps=1000)

        # run MD
        MaxwellBoltzmannDistribution(atoms, temp * units.kB)
        integrator = VelocityVerlet(atoms=atoms, timestep=self.time_step * units.fs)
        logger = MDLogger(
            dyn=integrator,
            atoms=atoms,
            logfile=str(
                Path(self.job_config.metadata.results_dir)
                / f"thermo_{num_jobs}-{job_num}.log"
            ),
            peratom=True,
        )
        integrator.attach(logger, interval=self.save_frequency)

        for _step in tqdm(range(self.steps)):
            integrator.run(1)

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        """Write calculation results to a compressed JSON file.

        Args:
            results: List of dictionaries containing energy and forces results
            results_dir: Directory path where results will be saved
            job_num: Index of the current job
            num_jobs: Total number of jobs
        """
        assert (
            Path(self.job_config.metadata.results_dir)
            / f"thermo_{num_jobs}-{job_num}.log"
        ).exists()
        time, Et = get_thermo(
            str(
                Path(self.job_config.metadata.results_dir)
                / f"thermo_{num_jobs}-{job_num}.log"
            )
        )
        assert len(time) == len(Et)

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



"""
Copyright (c) Meta, Inc. and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import gzip
import json
import os
from glob import glob
from typing import Callable, TypeVar

import pandas as pd

from fairchem.core.components.benchmark.benchmark_reducer import (
    JsonDFReducer,
)

R = TypeVar("R")
M = TypeVar("M")


class OMolReducer(JsonDFReducer):
    def __init__(
        self,
        benchmark_name: str,
        evaluator: Callable | None = None,
        benchmark_labels: str | None = None,
    ):
        """
        Args:
            benchmark_name: Name of the benchmark to run + evaluate
            evaluator: Evaluation code
            benchmark_labels: Benchmark labels to run through evaluator
        """
        self.benchmark_name = benchmark_name
        self.benchmark_labels = benchmark_labels
        self.evaluator = evaluator

    def join_results(self, results_dir: str, glob_pattern: str) -> pd.DataFrame:
        """Join results from multiple JSON files into a single DataFrame.

        Args:
            results_dir: Directory containing result files
            glob_pattern: Pattern to match result files

        Returns:
            Combined DataFrame containing all results
        """
        results = {}
        for filepath in glob(os.path.join(results_dir, glob_pattern)):
            with gzip.open(filepath, "rt") as f:
                data = json.load(f)
                results.update(data)
        return results

    def save_results(self, results: pd.DataFrame, results_dir: str) -> None:
        """Save joined results to a compressed json file

        Args:
            results:  results: Combined results from join_results
            results_dir: Directory containing result files
        """
        results_path = os.path.join(
            results_dir, f"{self.benchmark_name}_results.json.gz"
        )
        with gzip.open(results_path, "wb") as f:
            f.write(json.dumps(results).encode("utf-8"))

    def compute_metrics(self, results: dict, run_name: str) -> pd.DataFrame:
        if self.benchmark_labels and self.evaluator:
            with open(self.benchmark_labels) as f:
                target = json.load(f)
            metrics = self.evaluator(target, results)
            return pd.DataFrame([metrics], index=[run_name])
        return None



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import functools
import logging
import os
import random
import timeit
import uuid
from collections import defaultdict

import numpy as np
import torch
from ase import build
from torch.profiler import ProfilerActivity, profile

from fairchem.core.common.profiler_utils import get_profile_schedule
from fairchem.core.components.runner import Runner
from fairchem.core.datasets import data_list_collater
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.units.mlip_unit import MLIPPredictUnit
from fairchem.core.units.mlip_unit.api.inference import (
    InferenceSettings,
    inference_settings_default,
)


def seed_everywhere(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def ase_to_graph(atoms, neighbors: int, cutoff: float, external_graph=True):
    data_object = AtomicData.from_ase(
        atoms,
        max_neigh=neighbors,
        radius=cutoff,
        r_edges=external_graph,
    )
    data_object.natoms = torch.tensor(len(atoms))
    data_object.charge = torch.LongTensor([0])
    data_object.spin = torch.LongTensor([0])
    data_object.dataset = "omat"
    data_object.pos.requires_grad = True
    data_loader = torch.utils.data.DataLoader(
        [data_object],
        collate_fn=data_list_collater,
        batch_size=1,
        shuffle=False,
    )
    return next(iter(data_loader))


def get_fcc_carbon_xtal(
    neighbors: int,
    radius: float,
    num_atoms: int,
    lattice_constant: float = 3.8,
    external_graph: bool = True,
):
    # lattice_constant = 3.8, fcc generates a supercell with ~50 edges/atom
    atoms = build.bulk("C", "fcc", a=lattice_constant)
    n_cells = int(np.ceil(np.cbrt(num_atoms)))
    atoms = atoms.repeat((n_cells, n_cells, n_cells))
    indices = np.random.choice(len(atoms), num_atoms, replace=False)
    sampled_atoms = atoms[indices]
    return ase_to_graph(sampled_atoms, neighbors, radius, external_graph)


def get_qps(data, predictor, warmups: int = 10, timeiters: int = 100):
    def timefunc():
        predictor.predict(data)
        torch.cuda.synchronize()

    for _ in range(warmups):
        timefunc()
        logging.info(f"memory allocated: {torch.cuda.memory_allocated()/(1024**3)}")

    result = timeit.timeit(timefunc, number=timeiters)
    qps = timeiters / result
    ns_per_day = qps * 24 * 3600 / 1e6
    return qps, ns_per_day


def trace_handler(p, name, save_loc):
    trace_name = f"{name}.pt.trace.json"
    output_path = os.path.join(save_loc, trace_name)
    logging.info(f"Saving trace in {output_path}")
    p.export_chrome_trace(output_path)


def make_profile(data, predictor, name, save_loc):
    activities = [ProfilerActivity.CPU, ProfilerActivity.CUDA]
    profile_schedule, total_profile_steps = get_profile_schedule()
    tc = functools.partial(trace_handler, name=name, save_loc=save_loc)

    with profile(
        activities=activities,
        schedule=profile_schedule,
        on_trace_ready=tc,
    ) as p:
        for _ in range(total_profile_steps):
            predictor.predict(data)
            torch.cuda.synchronize()
            p.step()


class InferenceBenchRunner(Runner):
    def __init__(
        self,
        run_dir_root,
        natoms_list: list[int],
        model_checkpoints: dict[str, str],
        timeiters: int = 10,
        seed: int = 1,
        device="cuda",
        overrides: dict | None = None,
        inference_settings: InferenceSettings = inference_settings_default(),  # noqa B008
        generate_traces: bool = False,  # takes additional memory and time
    ):
        self.natoms_list = natoms_list
        self.device = device
        self.seed = seed
        self.timeiters = timeiters
        self.model_checkpoints = model_checkpoints
        self.run_dir = os.path.join(run_dir_root, uuid.uuid4().hex.upper()[0:8])
        self.overrides = overrides
        self.inference_settings = inference_settings
        self.generate_traces = generate_traces
        os.makedirs(self.run_dir, exist_ok=True)

    def run(self) -> None:
        seed_everywhere(self.seed)

        model_to_qps_data = defaultdict(list)

        for name, model_checkpoint in self.model_checkpoints.items():
            logging.info(
                f"Loading model: {model_checkpoint}, inference_settings: {self.inference_settings}"
            )
            predictor = MLIPPredictUnit(
                model_checkpoint,
                self.device,
                overrides=self.overrides,
                inference_settings=self.inference_settings,
            )
            max_neighbors = predictor.model.module.backbone.max_neighbors
            cutoff = predictor.model.module.backbone.cutoff
            logging.info(f"Model's max_neighbors: {max_neighbors}, cutoff: {cutoff}")

            # benchmark all cell sizes
            for natoms in self.natoms_list:
                data = get_fcc_carbon_xtal(
                    max_neighbors,
                    cutoff,
                    natoms,
                    external_graph=self.inference_settings.external_graph_gen,
                )
                num_atoms = data.natoms.item()

                print_info = f"Starting profile: model: {model_checkpoint}, num_atoms: {num_atoms}"
                if self.inference_settings.external_graph_gen:
                    num_edges = data.edge_index.shape[1]
                    print_info += f" num edges compute on: {num_edges}"
                logging.info(print_info)
                if self.generate_traces:
                    make_profile(data, predictor, name=name, save_loc=self.run_dir)
                qps, ns_per_day = get_qps(data, predictor, timeiters=self.timeiters)
                model_to_qps_data[name].append([num_atoms, ns_per_day])
                logging.info(
                    f"Profile results: model: {model_checkpoint}, num_atoms: {num_atoms}, qps: {qps}, ns_per_day: {ns_per_day}"
                )

    def save_state(self, _):
        return

    def load_state(self, _):
        return



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal

import torch

from fairchem.core.units.mlip_unit.api.inference import (
    InferenceSettings,
    guess_inference_settings,
)
from fairchem.core.units.mlip_unit.predict import MLIPPredictUnit

if TYPE_CHECKING:
    from pathlib import Path


def load_predict_unit(
    path: str | Path,
    inference_settings: InferenceSettings | str = "default",
    overrides: dict | None = None,
    device: Literal["cuda", "cpu"] | None = None,
    atom_refs: dict | None = None,
) -> MLIPPredictUnit:
    """Load a MLIPPredictUnit from a checkpoint file.

    Args:
        path: Path to the checkpoint file
        inference_settings: Settings for inference. Can be "default" (general purpose) or "turbo"
            (optimized for speed but requires fixed atomic composition). Advanced use cases can
            use a custom InferenceSettings object.
        overrides: Optional dictionary of settings to override default inference settings.
        device: Optional torch device to load the model onto.
        atom_refs: Optional dictionary of isolated atom reference energies.

    Returns:
        A MLIPPredictUnit instance ready for inference
    """

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        logging.warning(f"device was not explicitly set, using {device=}.")

    inference_settings = guess_inference_settings(inference_settings)
    overrides = overrides or {"backbone": {"always_use_pbc": False}}

    return MLIPPredictUnit(
        path,
        device=device,
        inference_settings=inference_settings,
        overrides=overrides,
        atom_refs=atom_refs,
    )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
import shutil
from typing import TYPE_CHECKING, Optional, Protocol, Union, runtime_checkable

from torchtnt.framework.callback import Callback
from torchtnt.framework.fit import fit

from fairchem.core.common import distutils
from fairchem.core.common.utils import get_subdirectories_sorted_by_time
from fairchem.core.components.runner import Runner

if TYPE_CHECKING:
    import torch
    from torchtnt.framework import EvalUnit, TrainUnit
    from torchtnt.framework.state import State
    from torchtnt.framework.unit import TTrainUnit


@runtime_checkable
class Checkpointable(Protocol):
    """
    Protocol that Units used by this trainer should implement if they want save and resume functionality
    This is in addition to Pytorch's Stateful protocol because it allows units implement custom logic
    that's required for checkpointing
    """

    def save_state(self, checkpoint_location: str) -> None:
        """
        Save the unit state to a checkpoint path

        Args:
            checkpoint_location: The checkpoint path to save to
        """

        ...

    def load_state(self, checkpoint_location: str | None) -> None:
        """
        Loads the state given a checkpoint path

        Args:
            checkpoint_location: The checkpoint path to restore from
        """

        ...


def get_most_recent_viable_checkpoint_path(checkpoint_dir: str | None) -> str | None:
    if not checkpoint_dir:
        return None

    ckpt_dirs_time = get_subdirectories_sorted_by_time(checkpoint_dir)
    most_recent_viable_checkpoint = None
    for sub_dir_path, _ in ckpt_dirs_time[::-1]:
        items = os.listdir(sub_dir_path)
        if items and ".metadata" in items:
            most_recent_viable_checkpoint = sub_dir_path
            break
    return most_recent_viable_checkpoint


class TrainCheckpointCallback(Callback):
    def __init__(
        self,
        checkpoint_every_n_steps: int,
        max_saved_checkpoints: int = 2,
    ):
        self.checkpoint_every_n_steps = checkpoint_every_n_steps
        self.max_saved_checkpoints = max_saved_checkpoints
        self.save_callback = None
        self.load_callback = None
        self.checkpoint_dir = None

    def set_runner_callbacks(
        self, save_callback: callable, load_callback: callable, checkpoint_dir: str
    ) -> None:
        self.save_callback = save_callback
        self.load_callback = load_callback
        self.checkpoint_dir = checkpoint_dir

    def on_train_step_start(self, state: State, unit: TTrainUnit) -> None:
        # We try to save the checkpoint on_train_step_start instead of at the on_train_step_end because both the step and epoch counts are consistently updated before it gets here
        # if we did this at on_train_step_end, the step would be correct but the epoch would have not been incremented and break the edge case on the last step of an epoch
        assert (
            self.save_callback
        ), "Must initialize set_checkpoint_call_backs from Runner!"
        step = unit.train_progress.num_steps_completed
        if (
            self.checkpoint_every_n_steps is not None
            and step % self.checkpoint_every_n_steps == 0
        ):
            self.save_callback(os.path.join(self.checkpoint_dir, f"step_{step}"))
            # on main rank only
            # if there are too many checkpoints, delete the oldest one
            if distutils.is_master():
                checkpoint_dirs_by_time = get_subdirectories_sorted_by_time(
                    self.checkpoint_dir
                )
                for dir, _ in checkpoint_dirs_by_time[: -self.max_saved_checkpoints]:
                    if not os.path.islink(dir):
                        shutil.rmtree(dir)

    def on_train_end(self, state: State, unit: TTrainUnit) -> None:
        if self.checkpoint_every_n_steps is not None:
            # also always checkpoint on train end
            assert (
                self.save_callback
            ), "Must initialize set_checkpoint_call_backs from Runner!"
            self.save_callback(os.path.join(self.checkpoint_dir, "final"))


class TrainEvalRunner(Runner):
    def __init__(
        self,
        train_dataloader: torch.utils.data.dataloader,
        eval_dataloader: torch.utils.data.dataloader,
        train_eval_unit: Union[TrainUnit, EvalUnit, Checkpointable],
        callbacks: list[Callback] | None = None,
        max_epochs: int | None = 1,
        evaluate_every_n_steps: Optional[int] = None,
        max_steps: int | None = None,
    ):
        self.train_dataloader = train_dataloader
        self.eval_dataloader = eval_dataloader
        self.train_eval_unit = train_eval_unit
        self.callbacks = callbacks if callbacks is not None else []
        self.max_epochs = max_epochs
        self.max_steps = max_steps
        self.evaluate_every_n_steps = evaluate_every_n_steps

        checkpoint_callbacks = [
            c for c in callbacks if isinstance(c, TrainCheckpointCallback)
        ]
        assert len(checkpoint_callbacks) <= 1
        self.checkpoint_callback = (
            checkpoint_callbacks[0] if len(checkpoint_callbacks) == 1 else None
        )
        logging.info(f"Train Dataloader size {len(self.train_dataloader)}")
        logging.info(f"Eval Dataloader size {len(self.eval_dataloader)}")

    def run(self) -> None:
        if self.checkpoint_callback is not None:
            self.checkpoint_callback.set_runner_callbacks(
                self.save_state,
                self.load_state,
                self.job_config.metadata.checkpoint_dir,
            )

        fit(
            self.train_eval_unit,
            train_dataloader=self.train_dataloader,
            eval_dataloader=self.eval_dataloader,
            max_epochs=self.max_epochs,
            max_steps=self.max_steps,
            callbacks=self.callbacks,
            evaluate_every_n_steps=self.evaluate_every_n_steps,
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        # in the case of preemption, don't attempt to save a new checkpoint but try to move an existing to the checkpoint_location
        # this is because submitit's preemption routine only calls checkpoint on master and dcp will deadlock if its not called on all ranks
        if is_preemption:
            most_recent_checkpoint_path = get_most_recent_viable_checkpoint_path(
                self.job_config.metadata.checkpoint_dir
            )
            if most_recent_checkpoint_path:
                if os.path.lexists(checkpoint_location):
                    logging.warning(
                        f"Checkpoint location {checkpoint_location} already exists, removing it"
                    )
                    os.remove(checkpoint_location)
                os.symlink(most_recent_checkpoint_path, checkpoint_location)
                logging.info(
                    f"When the job resumes from preemption, it will be using the state found at {most_recent_checkpoint_path}, which has been symlinked to {checkpoint_location}"
                )
                return True
            else:
                logging.info(
                    "Did not find a viable checkpoint, no preemption checkpoint is available"
                )
                return False

        self.train_eval_unit.save_state(checkpoint_location)
        return True

    def load_state(self, checkpoint_location: str | None) -> None:
        # if checkpoint_location is given, load that, otherwise attempt to load from latest checkpoint
        if checkpoint_location:
            checkpoint_to_load = checkpoint_location
        else:
            # we could be here because of a node failure where a checkpoint exists but the preemption code was never triggered
            # in this case we attempt to find the last known checkpoint
            # NOTE we must do this otherwise an automatically requeue by the cluster could restart this job from step 0
            most_recent_checkpoint_path = get_most_recent_viable_checkpoint_path(
                self.job_config.metadata.checkpoint_dir
            )
            if most_recent_checkpoint_path:
                logging.info(
                    f"Last existing checkpoints found at {most_recent_checkpoint_path}, starting from here"
                )
                checkpoint_to_load = most_recent_checkpoint_path
            else:
                logging.info("No existing checkpoints found, starting from scratch")
                return

        self.train_eval_unit.load_state(checkpoint_to_load)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import torch
import torch.nn as nn
from torch.profiler import record_function
from typing_extensions import Literal

from fairchem.core.common import gp_utils
from fairchem.core.models.uma.nn.activation import (
    GateActivation,
    SeparableS2Activation_M,
)
from fairchem.core.models.uma.nn.layer_norm import (
    get_normalization_layer,
)
from fairchem.core.models.uma.nn.mole import MOLE
from fairchem.core.models.uma.nn.radial import PolynomialEnvelope
from fairchem.core.models.uma.nn.so2_layers import SO2_Convolution
from fairchem.core.models.uma.nn.so3_layers import SO3_Linear

if TYPE_CHECKING:
    from fairchem.core.models.uma.common.so3 import CoefficientMapping, SO3_Grid


def set_mole_ac_start_index(module: nn.Module, index: int) -> None:
    for submodule in module.modules():
        if isinstance(submodule, MOLE):
            submodule.global_mole_tensors.ac_start_idx = index


class Edgewise(torch.nn.Module):
    def __init__(
        self,
        sphere_channels: int,
        hidden_channels: int,
        lmax: int,
        mmax: int,
        edge_channels_list: list[int],
        mappingReduced: CoefficientMapping,
        SO3_grid: SO3_Grid,
        cutoff: float,
        # Enables activation checkpointing of edges in
        # activation_checkpoint_chunk_size size edge blocks
        activation_checkpoint_chunk_size: int | None,
        act_type: Literal["gate", "s2"] = "gate",
    ):
        super().__init__()

        self.sphere_channels = sphere_channels
        self.hidden_channels = hidden_channels
        self.lmax = lmax
        self.mmax = mmax
        self.activation_checkpoint_chunk_size = activation_checkpoint_chunk_size

        self.mappingReduced = mappingReduced
        self.SO3_grid = SO3_grid
        self.edge_channels_list = copy.deepcopy(edge_channels_list)
        self.act_type = act_type

        if self.act_type == "gate":
            self.act = GateActivation(
                lmax=self.lmax,
                mmax=self.mmax,
                num_channels=self.hidden_channels,
                m_prime=True,
            )
            extra_m0_output_channels = self.lmax * self.hidden_channels
        elif self.act_type == "s2":
            # NOTE: this is the only place where the SO3 grid of the edges (lmax/mmax) is used
            self.act = SeparableS2Activation_M(
                lmax=self.lmax,
                mmax=self.mmax,
                SO3_grid=self.SO3_grid,
                to_m=self.mappingReduced.to_m,
            )
            extra_m0_output_channels = self.hidden_channels
        else:
            raise ValueError(f"Unknown activation type {self.act_type}")

        self.so2_conv_1 = SO2_Convolution(
            2 * self.sphere_channels,
            self.hidden_channels,
            self.lmax,
            self.mmax,
            self.mappingReduced,
            internal_weights=False,
            edge_channels_list=self.edge_channels_list,
            extra_m0_output_channels=extra_m0_output_channels,
        )

        self.so2_conv_2 = SO2_Convolution(
            self.hidden_channels,
            self.sphere_channels,
            self.lmax,
            self.mmax,
            self.mappingReduced,
            internal_weights=True,
            edge_channels_list=None,
            extra_m0_output_channels=None,
        )

        self.cutoff = cutoff
        self.envelope = PolynomialEnvelope(exponent=5)

        self.out_mask = self.SO3_grid["lmax_lmax"].mapping.coefficient_idx(
            self.lmax, self.mmax
        )

    def forward(
        self,
        x,
        x_edge,
        edge_distance,
        edge_index,
        wigner_and_M_mapping,
        wigner_and_M_mapping_inv,
        node_offset: int = 0,
    ):
        # we perform the all gather upfront once during each forward call so we don't need to repeat this multiple times during activation checkpointing.
        if gp_utils.initialized():
            x_full = gp_utils.gather_from_model_parallel_region_sum_grad(x, dim=0)
        else:
            x_full = x

        if self.activation_checkpoint_chunk_size is None:
            return self.forward_chunk(
                x_full,
                x.shape[0],
                x_edge,
                edge_distance,
                edge_index,
                wigner_and_M_mapping,
                wigner_and_M_mapping_inv,
                node_offset,
            )
        edge_index_partitions = edge_index.split(
            self.activation_checkpoint_chunk_size, dim=1
        )
        wigner_partitions = wigner_and_M_mapping.split(
            self.activation_checkpoint_chunk_size, dim=0
        )
        wigner_inv_partitions = wigner_and_M_mapping_inv.split(
            self.activation_checkpoint_chunk_size, dim=0
        )
        edge_distance_parititons = edge_distance.split(
            self.activation_checkpoint_chunk_size, dim=0
        )
        x_edge_partitions = x_edge.split(self.activation_checkpoint_chunk_size, dim=0)
        new_embeddings = []
        # when chunking, we need to keep track of the start index of the chunk and give this information
        # to the mole layers
        ac_mole_start_idx = 0

        for idx in range(len(edge_index_partitions)):
            new_embeddings.append(
                torch.utils.checkpoint.checkpoint(
                    self.forward_chunk,
                    x_full,
                    x.shape[0],
                    x_edge_partitions[idx],
                    edge_distance_parititons[idx],
                    edge_index_partitions[idx],
                    wigner_partitions[idx],
                    wigner_inv_partitions[idx],
                    node_offset,
                    ac_mole_start_idx,
                    use_reentrant=False,
                )
            )
            ac_mole_start_idx += edge_index_partitions[idx].shape[1]

            if len(new_embeddings) > 8:
                new_embeddings = [torch.stack(new_embeddings).sum(axis=0)]
        return torch.stack(new_embeddings).sum(axis=0)

    def forward_chunk(
        self,
        x_full,
        x_original_shape,
        x_edge,
        edge_distance,
        edge_index,
        wigner_and_M_mapping,
        wigner_and_M_mapping_inv,
        node_offset: int = 0,
        ac_mole_start_idx: int = 0,
    ):
        # here we need to update the ac_start_idx of the mole layers under here for this chunking to
        # work properly with MoLE together
        set_mole_ac_start_index(self, ac_mole_start_idx)

        x_source = x_full[edge_index[0]]
        x_target = x_full[edge_index[1]]

        x_message = torch.cat((x_source, x_target), dim=2)

        with record_function("SO2Conv"):
            # Rotate the irreps to align with the edge
            x_message = torch.bmm(wigner_and_M_mapping, x_message)

            # SO2 convolution
            x_message, x_0_gating = self.so2_conv_1(x_message, x_edge)

            # M-prime...
            x_message = self.act(x_0_gating, x_message)

            x_message = self.so2_conv_2(x_message, x_edge)

            # envelope
            dist_scaled = edge_distance / self.cutoff
            env = self.envelope(dist_scaled)
            x_message = x_message * env.view(-1, 1, 1)

            # Rotate back the irreps
            x_message = torch.bmm(wigner_and_M_mapping_inv, x_message)

        # Compute the sum of the incoming neighboring messages for each target node
        new_embedding = torch.zeros(
            (x_original_shape,) + x_message.shape[1:],
            dtype=x_message.dtype,
            device=x_message.device,
        )

        new_embedding.index_add_(0, edge_index[1] - node_offset, x_message)
        # reset ac start index
        set_mole_ac_start_index(self, 0)
        return new_embedding


class SpectralAtomwise(torch.nn.Module):
    def __init__(
        self,
        sphere_channels: int,
        hidden_channels: int,
        lmax: int,
        mmax: int,
        SO3_grid: SO3_Grid,
    ):
        super().__init__()
        self.sphere_channels = sphere_channels
        self.hidden_channels = hidden_channels
        self.lmax = lmax
        self.mmax = mmax
        self.SO3_grid = SO3_grid

        self.scalar_mlp = nn.Sequential(
            nn.Linear(
                self.sphere_channels,
                self.lmax * self.hidden_channels,
                bias=True,
            ),
            nn.SiLU(),
        )

        self.so3_linear_1 = SO3_Linear(
            self.sphere_channels, self.hidden_channels, lmax=self.lmax
        )
        self.act = GateActivation(
            lmax=self.lmax, mmax=self.lmax, num_channels=self.hidden_channels
        )
        self.so3_linear_2 = SO3_Linear(
            self.hidden_channels, self.sphere_channels, lmax=self.lmax
        )

    def forward(self, x):
        gating_scalars = self.scalar_mlp(x.narrow(1, 0, 1))
        x = self.so3_linear_1(x)
        x = self.act(gating_scalars, x)
        x = self.so3_linear_2(x)
        return x


class GridAtomwise(torch.nn.Module):
    def __init__(
        self,
        sphere_channels: int,
        hidden_channels: int,
        lmax: int,
        mmax: int,
        SO3_grid: SO3_Grid,
    ):
        super().__init__()
        self.sphere_channels = sphere_channels
        self.hidden_channels = hidden_channels
        self.lmax = lmax
        self.mmax = mmax
        self.SO3_grid = SO3_grid

        self.grid_mlp = nn.Sequential(
            nn.Linear(self.sphere_channels, self.hidden_channels, bias=False),
            nn.SiLU(),
            nn.Linear(self.hidden_channels, self.hidden_channels, bias=False),
            nn.SiLU(),
            nn.Linear(self.hidden_channels, self.sphere_channels, bias=False),
        )

    def forward(self, x):
        # Project to grid
        x_grid = self.SO3_grid["lmax_lmax"].to_grid(x, self.lmax, self.lmax)
        # Perform point-wise operations
        x_grid = self.grid_mlp(x_grid)
        # Project back to spherical harmonic coefficients
        x = self.SO3_grid["lmax_lmax"].from_grid(x_grid, self.lmax, self.lmax)
        return x


class eSCNMD_Block(torch.nn.Module):
    def __init__(
        self,
        sphere_channels: int,
        hidden_channels: int,
        lmax: int,
        mmax: int,
        mappingReduced: CoefficientMapping,
        SO3_grid: SO3_Grid,
        edge_channels_list: list[int],
        cutoff: float,
        norm_type: Literal["layer_norm", "layer_norm_sh", "rms_norm_sh"],
        act_type: Literal["gate", "s2"],
        ff_type: Literal["spectral", "grid"],
        activation_checkpoint_chunk_size: int | None,
    ) -> None:
        super().__init__()
        self.sphere_channels = sphere_channels
        self.hidden_channels = hidden_channels
        self.lmax = lmax
        self.mmax = mmax

        self.norm_1 = get_normalization_layer(
            norm_type, lmax=self.lmax, num_channels=sphere_channels
        )

        self.edge_wise = Edgewise(
            sphere_channels=sphere_channels,
            hidden_channels=hidden_channels,
            lmax=lmax,
            mmax=mmax,
            edge_channels_list=edge_channels_list,
            mappingReduced=mappingReduced,
            SO3_grid=SO3_grid,
            cutoff=cutoff,
            act_type=act_type,
            activation_checkpoint_chunk_size=activation_checkpoint_chunk_size,
        )

        self.norm_2 = get_normalization_layer(
            norm_type, lmax=self.lmax, num_channels=sphere_channels
        )

        if ff_type == "spectral":
            self.atom_wise = SpectralAtomwise(
                sphere_channels=sphere_channels,
                hidden_channels=hidden_channels,
                lmax=lmax,
                mmax=mmax,
                SO3_grid=SO3_grid,
            )
        elif ff_type == "grid":
            self.atom_wise = GridAtomwise(
                sphere_channels=sphere_channels,
                hidden_channels=hidden_channels,
                lmax=lmax,
                mmax=mmax,
                SO3_grid=SO3_grid,
            )

    def forward(
        self,
        x,
        x_edge,
        edge_distance,
        edge_index,
        wigner_and_M_mapping,
        wigner_and_M_mapping_inv,
        sys_node_embedding=None,
        node_offset: int = 0,
    ):
        x_res = x
        x = self.norm_1(x)

        if sys_node_embedding is not None:
            x[:, 0, :] = x[:, 0, :] + sys_node_embedding

        with record_function("edgewise"):
            x = self.edge_wise(
                x,
                x_edge,
                edge_distance,
                edge_index,
                wigner_and_M_mapping,
                wigner_and_M_mapping_inv,
                node_offset,
            )
            x = x + x_res

        x_res = x
        x = self.norm_2(x)

        with record_function("atomwise"):
            x = self.atom_wise(x)
            x = x + x_res

        return x



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Copied (and improved) from:
https://github.com/fadel/pytorch_ema/blob/master/torch_ema/ema.py (MIT license)
"""

from __future__ import annotations

import copy
import weakref
from typing import TYPE_CHECKING

import torch

from fairchem.core.common.typing import none_throws

if TYPE_CHECKING:
    from collections.abc import Iterable


# Partially based on:
# https://github.com/tensorflow/tensorflow/blob/r1.13/tensorflow/python/training/moving_averages.py
class ExponentialMovingAverage:
    """
    Maintains (exponential) moving average of a set of parameters.

    Args:
        parameters: Iterable of `torch.nn.Parameter` (typically from
            `model.parameters()`).
        decay: The exponential decay.
        use_num_updates: Whether to use number of updates when computing
            averages.
    """

    def __init__(
        self,
        parameters: Iterable[torch.nn.Parameter],
        decay: float,
        use_num_updates: bool = False,
    ) -> None:
        if decay < 0.0 or decay > 1.0:
            raise ValueError("Decay must be between 0 and 1")
        self.decay = decay
        self.num_updates: int | None = 0 if use_num_updates else None
        parameters = list(parameters)
        self.shadow_params = [p.clone().detach() for p in parameters if p.requires_grad]
        self.collected_params: list[torch.nn.Parameter] = []
        # By maintaining only a weakref to each parameter,
        # we maintain the old GC behaviour of ExponentialMovingAverage:
        # if the model goes out of scope but the ExponentialMovingAverage
        # is kept, no references to the model or its parameters will be
        # maintained, and the model will be cleaned up.
        self._params_refs = [weakref.ref(p) for p in parameters if p.requires_grad]

    def _get_parameters(
        self, parameters: Iterable[torch.nn.Parameter] | None
    ) -> Iterable[torch.nn.Parameter]:
        none_msg = (
            "(One of) the parameters with which this "
            "ExponentialMovingAverage "
            "was initialized no longer exists (was garbage collected);"
            " please either provide `parameters` explicitly or keep "
            "the model to which they belong from being garbage "
            "collected."
        )
        if parameters is None:
            return [none_throws(p(), none_msg) for p in self._params_refs]
        else:
            return [p for p in parameters if p.requires_grad]

    def update(self, parameters: Iterable[torch.nn.Parameter] | None = None) -> None:
        """
        Update currently maintained parameters.

        Call this every time the parameters are updated, such as the result of
        the `optimizer.step()` call.

        Args:
          parameters: Iterable of `torch.nn.Parameter`; usually the same set of
            parameters used to initialize this object. If `None`, the
            parameters with which this `ExponentialMovingAverage` was
            initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        decay = self.decay
        if self.num_updates is not None:
            self.num_updates += 1
            decay = min(decay, (1 + self.num_updates) / (10 + self.num_updates))
        one_minus_decay = 1.0 - decay
        with torch.no_grad():
            for s_param, param in zip(self.shadow_params, parameters):
                tmp = param - s_param
                s_param.add_(tmp, alpha=one_minus_decay)

    def copy_to(self, parameters: Iterable[torch.nn.Parameter] | None = None) -> None:
        """
        Copy current parameters into given collection of parameters.

        Args:
          parameters: Iterable of `torch.nn.Parameter`; the parameters to be
            updated with the stored moving averages. If `None`, the
            parameters with which this `ExponentialMovingAverage` was
            initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        for s_param, param in zip(self.shadow_params, parameters):
            param.data.copy_(s_param.data)

    def store(self, parameters: Iterable[torch.nn.Parameter] | None = None) -> None:
        """
        Save the current parameters for restoring later.

        Args:
          parameters: Iterable of `torch.nn.Parameter`; the parameters to be
            temporarily stored. If `None`, the parameters of with which this
            `ExponentialMovingAverage` was initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        self.collected_params = [param.clone() for param in parameters]

    def restore(self, parameters: Iterable[torch.nn.Parameter] | None = None) -> None:
        """
        Restore the parameters stored with the `store` method.
        Useful to validate the model with EMA parameters without affecting the
        original optimization process. Store the parameters before the
        `copy_to` method. After validation (or model saving), use this to
        restore the former parameters.

        Args:
          parameters: Iterable of `torch.nn.Parameter`; the parameters to be
            updated with the stored parameters. If `None`, the
            parameters with which this `ExponentialMovingAverage` was
            initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        for c_param, param in zip(self.collected_params, parameters):
            param.data.copy_(c_param.data)

    def state_dict(self) -> dict:
        r"""Returns the state of the ExponentialMovingAverage as a dict."""
        # Following PyTorch conventions, references to tensors are returned:
        # "returns a reference to the state and not its copy!" -
        # https://pytorch.org/tutorials/beginner/saving_loading_models.html#what-is-a-state-dict
        return {
            "decay": self.decay,
            "num_updates": self.num_updates,
            "shadow_params": self.shadow_params,
            "collected_params": self.collected_params,
        }

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the ExponentialMovingAverage state.

        Args:
            state_dict (dict): EMA state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        # deepcopy, to be consistent with module API
        state_dict = copy.deepcopy(state_dict)

        self.decay = state_dict["decay"]
        if self.decay < 0.0 or self.decay > 1.0:
            raise ValueError("Decay must be between 0 and 1")

        self.num_updates = state_dict["num_updates"]
        assert self.num_updates is None or isinstance(
            self.num_updates, int
        ), "Invalid num_updates"

        assert isinstance(
            state_dict["shadow_params"], list
        ), "shadow_params must be a list"
        self.shadow_params = [
            p.to(self.shadow_params[i].device)
            for i, p in enumerate(state_dict["shadow_params"])
        ]
        assert all(
            isinstance(p, torch.Tensor) for p in self.shadow_params
        ), "shadow_params must all be Tensors"

        assert isinstance(
            state_dict["collected_params"], list
        ), "collected_params must be a list"
        # collected_params is empty at initialization,
        # so use shadow_params for device instead
        self.collected_params = [
            p.to(self.shadow_params[i].device)
            for i, p in enumerate(state_dict["collected_params"])
        ]
        assert all(
            isinstance(p, torch.Tensor) for p in self.collected_params
        ), "collected_params must all be Tensors"



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import torch

from fairchem.core.graph.radius_graph_pbc import radius_graph_pbc, radius_graph_pbc_v2


def get_pbc_distances(
    pos,
    edge_index,
    cell,
    cell_offsets,
    neighbors,
    return_offsets: bool = False,
    return_distance_vec: bool = False,
):
    row, col = edge_index

    distance_vectors = pos[row] - pos[col]

    # correct for pbc
    neighbors = neighbors.to(cell.device)
    cell = torch.repeat_interleave(cell, neighbors, dim=0)
    offsets = cell_offsets.float().view(-1, 1, 3).bmm(cell.float()).view(-1, 3)
    distance_vectors += offsets

    # compute distances
    distances = distance_vectors.norm(dim=-1)

    # redundancy: remove zero distances
    nonzero_idx = torch.arange(len(distances), device=distances.device)[distances != 0]
    edge_index = edge_index[:, nonzero_idx]
    distances = distances[nonzero_idx]

    out = {
        "edge_index": edge_index,
        "distances": distances,
    }

    if return_distance_vec:
        out["distance_vec"] = distance_vectors[nonzero_idx]

    if return_offsets:
        out["offsets"] = offsets[nonzero_idx]

    return out


# TODO: compiling internal graph gen is not supported right now
@torch.compiler.disable()
def generate_graph(
    data: dict,  # this is still a torch geometric batch object currently, turn this into a dict
    cutoff: float,
    max_neighbors: int,
    enforce_max_neighbors_strictly: bool,
    radius_pbc_version: int,
    pbc: torch.Tensor,
) -> dict:
    """Generate a graph representation from atomic structure data.

    Args:
        data (dict): A dictionary containing a batch of molecular structures.
            It should have the following keys:
                - 'pos' (torch.Tensor): Positions of the atoms.
                - 'cell' (torch.Tensor): Cell vectors of the molecular structures.
                - 'natoms' (torch.Tensor): Number of atoms in each molecular structure.
        cutoff (float): The maximum distance between atoms to consider them as neighbors.
        max_neighbors (int): The maximum number of neighbors to consider for each atom.
        enforce_max_neighbors_strictly (bool): Whether to strictly enforce the maximum number of neighbors.
        radius_pbc_version: the version of radius_pbc impl
        pbc (list[bool]): The periodic boundary conditions in 3 dimensions, defaults to [True,True,True] for 3D pbc

    Returns:
        dict: A dictionary containing the generated graph with the following keys:
            - 'edge_index' (torch.Tensor): Indices of the edges in the graph.
            - 'edge_distance' (torch.Tensor): Distances between the atoms connected by the edges.
            - 'edge_distance_vec' (torch.Tensor): Vectors representing the distances between the atoms connected by the edges.
            - 'cell_offsets' (torch.Tensor): Offsets of the cell vectors for each edge.
            - 'offset_distances' (torch.Tensor): Distances between the atoms connected by the edges, including the cell offsets.
            - 'neighbors' (torch.Tensor): Number of neighbors for each atom.
    """

    if radius_pbc_version == 1:
        radius_graph_pbc_fn = radius_graph_pbc
    elif radius_pbc_version == 2:
        radius_graph_pbc_fn = radius_graph_pbc_v2
    else:
        raise ValueError(f"Invalid radius_pbc version {radius_pbc_version}")

    (
        edge_index_per_system,
        cell_offsets_per_system,
        neighbors_per_system,
    ) = list(
        zip(
            *[
                radius_graph_pbc_fn(
                    data[idx],  # loop over the batches?
                    cutoff,
                    max_neighbors,
                    enforce_max_neighbors_strictly,
                    pbc=pbc[idx],
                )
                for idx in range(len(data))
            ]
        )
    )

    # atom indexs in the edge_index need to be offset
    atom_index_offset = data.natoms.cumsum(dim=0).roll(1)
    atom_index_offset[0] = 0
    edge_index = torch.hstack(
        [
            edge_index_per_system[idx] + atom_index_offset[idx]
            for idx in range(len(data))
        ]
    )
    cell_offsets = torch.vstack(cell_offsets_per_system)
    neighbors = torch.hstack(neighbors_per_system)

    out = get_pbc_distances(
        data.pos,
        edge_index,
        data.cell,
        cell_offsets,
        neighbors,
        return_offsets=True,
        return_distance_vec=True,
    )

    edge_index = out["edge_index"]
    edge_dist = out["distances"]
    cell_offset_distances = out["offsets"]
    distance_vec = out["distance_vec"]

    return {
        "edge_index": edge_index,
        "edge_distance": edge_dist,
        "edge_distance_vec": distance_vec,
        "cell_offsets": cell_offsets,
        "offset_distances": cell_offset_distances,
        "neighbors": neighbors,
    }



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import itertools
import logging
import math
from contextlib import contextmanager
from typing import Callable, TypedDict

import torch
import torch.nn as nn


class _Stats(TypedDict):
    variance_in: float
    variance_out: float
    n_samples: int


IndexFn = Callable[[], None]


def _check_consistency(old: torch.Tensor, new: torch.Tensor, key: str) -> None:
    if not torch.allclose(old, new):
        raise ValueError(
            f"Scale factor parameter {key} is inconsistent with the loaded state dict.\n"
            f"Old: {old}\n"
            f"Actual: {new}"
        )


class ScaleFactor(nn.Module):
    scale_factor: torch.Tensor

    name: str | None = None
    index_fn: IndexFn | None = None
    stats: _Stats | None = None

    def __init__(
        self,
        name: str | None = None,
        enforce_consistency: bool = True,
    ) -> None:
        super().__init__()

        self.name = name
        self.index_fn = None
        self.stats = None

        self.scale_factor = nn.parameter.Parameter(
            torch.tensor(0.0), requires_grad=False
        )
        if enforce_consistency:
            self._register_load_state_dict_pre_hook(self._enforce_consistency)

    def _enforce_consistency(
        self,
        state_dict,
        prefix,
        _local_metadata,
        _strict,
        _missing_keys,
        _unexpected_keys,
        _error_msgs,
    ) -> None:
        if not self.fitted:
            return

        persistent_buffers = {
            k: v
            for k, v in self._buffers.items()
            if k not in self._non_persistent_buffers_set
        }
        local_name_params = itertools.chain(
            self._parameters.items(), persistent_buffers.items()
        )
        local_state = {k: v for k, v in local_name_params if v is not None}

        for name, param in local_state.items():
            key = prefix + name
            if key not in state_dict:
                continue

            input_param = state_dict[key]
            _check_consistency(old=param, new=input_param, key=key)

    @property
    def fitted(self) -> bool:
        return bool((self.scale_factor != 0.0).item())

    @torch.jit.unused
    def reset_(self) -> None:
        self.scale_factor.zero_()

    @torch.jit.unused
    def set_(self, scale: float | torch.Tensor) -> None:
        if self.fitted:
            _check_consistency(
                old=self.scale_factor,
                new=torch.tensor(scale) if isinstance(scale, float) else scale,
                key="scale_factor",
            )
        self.scale_factor.fill_(scale)

    @torch.jit.unused
    def initialize_(self, *, index_fn: IndexFn | None = None) -> None:
        self.index_fn = index_fn

    @contextmanager
    @torch.jit.unused
    def fit_context_(self):
        self.stats = _Stats(variance_in=0.0, variance_out=0.0, n_samples=0)
        yield
        del self.stats
        self.stats = None

    @torch.jit.unused
    def fit_(self):
        assert self.stats, "Stats not set"
        for k, v in self.stats.items():
            assert v > 0, f"{k} is {v}"

        self.stats["variance_in"] = self.stats["variance_in"] / self.stats["n_samples"]
        self.stats["variance_out"] = (
            self.stats["variance_out"] / self.stats["n_samples"]
        )

        ratio = self.stats["variance_out"] / self.stats["variance_in"]
        value = math.sqrt(1 / ratio)

        self.set_(value)

        stats = dict(**self.stats)
        return stats, ratio, value

    @torch.no_grad()
    @torch.jit.unused
    def _observe(self, x: torch.Tensor, ref: torch.Tensor | None = None) -> None:
        if self.stats is None:
            logging.debug("Observer not initialized but self.observe() called")
            return

        n_samples = x.shape[0]
        self.stats["variance_out"] += torch.mean(torch.var(x, dim=0)).item() * n_samples

        if ref is None:
            self.stats["variance_in"] += n_samples
        else:
            self.stats["variance_in"] += (
                torch.mean(torch.var(ref, dim=0)).item() * n_samples
            )
        self.stats["n_samples"] += n_samples

    def forward(
        self,
        x: torch.Tensor,
        *,
        ref: torch.Tensor | None = None,
    ) -> torch.Tensor:
        if self.index_fn is not None:
            self.index_fn()

        if self.fitted:
            x = x * self.scale_factor

        if not torch.jit.is_scripting():
            self._observe(x, ref=ref)

        return x



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import contextlib
import logging
from typing import Any

import numpy as np
import torch
from torch import distributed as dist
from torch.distributed.nn.functional import all_reduce

"""
Functions to support graph parallel training.
This is based on the Megatron-LM implementation:
https://github.com/facebookresearch/fairscale/blob/main/fairscale/nn/model_parallel/initialize.py
"""

########## INITIALIZATION ##########

_GRAPH_PARALLEL_GROUP = None
_DATA_PARALLEL_GROUP = None


def ensure_div(a: int, b: int) -> None:
    assert a % b == 0


def divide_and_check_no_remainder(a: int, b: int) -> int:
    ensure_div(a, b)
    return a // b


def setup_graph_parallel_groups(
    graph_parallel_group_size: int, distributed_backend: str
) -> None:
    assert torch.distributed.is_initialized()
    world_size = torch.distributed.get_world_size()
    assert (
        graph_parallel_group_size <= world_size
    ), "graph parallel group size must be at most world size"

    ensure_div(world_size, graph_parallel_group_size)
    dp_size = world_size // graph_parallel_group_size
    rank = dist.get_rank()

    if rank == 0:
        logging.info(
            f"> initializing graph parallel with size {graph_parallel_group_size}"
        )
        logging.info(f"> initializing ddp with size {dp_size}")

    groups = torch.arange(world_size).reshape(dp_size, graph_parallel_group_size)
    found = [x.item() for x in torch.where(groups == rank)]

    global _DATA_PARALLEL_GROUP
    assert _DATA_PARALLEL_GROUP is None, "data parallel group is already initialized"
    for j in range(graph_parallel_group_size):
        group = dist.new_group(groups[:, j].tolist(), backend=distributed_backend)
        if j == found[1]:
            _DATA_PARALLEL_GROUP = group
    global _GRAPH_PARALLEL_GROUP
    assert _GRAPH_PARALLEL_GROUP is None, "graph parallel group is already initialized"
    for i in range(dp_size):
        group = dist.new_group(groups[i, :].tolist(), backend=distributed_backend)
        if i == found[0]:
            _GRAPH_PARALLEL_GROUP = group


def setup_gp(config) -> None:
    gp_size = config["gp_gpus"]
    backend = config["distributed_backend"]
    assert torch.distributed.is_initialized()
    world_size = torch.distributed.get_world_size()

    gp_size = min(gp_size, world_size)
    ensure_div(world_size, gp_size)
    dp_size = world_size // gp_size
    rank = dist.get_rank()

    if rank == 0:
        logging.info(f"> initializing graph parallel with size {gp_size}")
        logging.info(f"> initializing ddp with size {dp_size}")

    groups = torch.arange(world_size).reshape(dp_size, gp_size)
    found = [x.item() for x in torch.where(groups == rank)]

    global _DATA_PARALLEL_GROUP
    assert _DATA_PARALLEL_GROUP is None, "data parallel group is already initialized"
    for j in range(gp_size):
        group = dist.new_group(groups[:, j].tolist(), backend=backend)
        if j == found[1]:
            _DATA_PARALLEL_GROUP = group
    global _GRAPH_PARALLEL_GROUP
    assert _GRAPH_PARALLEL_GROUP is None, "graph parallel group is already initialized"
    for i in range(dp_size):
        group = dist.new_group(groups[i, :].tolist(), backend=backend)
        if i == found[0]:
            _GRAPH_PARALLEL_GROUP = group


def cleanup_gp() -> None:
    global _DATA_PARALLEL_GROUP
    global _GRAPH_PARALLEL_GROUP
    assert _GRAPH_PARALLEL_GROUP is not None
    assert _DATA_PARALLEL_GROUP is not None
    with contextlib.suppress(ValueError):
        dist.destroy_process_group(_DATA_PARALLEL_GROUP)
    with contextlib.suppress(ValueError):
        dist.destroy_process_group(_GRAPH_PARALLEL_GROUP)
    _DATA_PARALLEL_GROUP = None
    _GRAPH_PARALLEL_GROUP = None


def initialized() -> bool:
    return _GRAPH_PARALLEL_GROUP is not None


def get_dp_group():
    return _DATA_PARALLEL_GROUP


def get_gp_group():
    return _GRAPH_PARALLEL_GROUP


def get_dp_rank() -> int:
    return dist.get_rank(group=get_dp_group())


def get_gp_rank() -> int:
    return dist.get_rank(group=get_gp_group())


def get_dp_world_size() -> int:
    return dist.get_world_size(group=get_dp_group())


def get_gp_world_size() -> int:
    return 1 if not initialized() else dist.get_world_size(group=get_gp_group())


########## DIST METHODS ##########


def pad_tensor(
    tensor: torch.Tensor, dim: int = -1, target_size: int | None = None
) -> torch.Tensor:
    size = tensor.size(dim)
    if target_size is None:
        world_size = get_gp_world_size()
        pad_size = 0 if size % world_size == 0 else world_size - size % world_size
    else:
        pad_size = target_size - size
    if pad_size == 0:
        return tensor
    pad_shape = list(tensor.shape)
    pad_shape[dim] = pad_size
    padding = torch.empty(pad_shape, device=tensor.device, dtype=tensor.dtype)
    return torch.cat([tensor, padding], dim=dim)


def trim_tensor(tensor: torch.Tensor, sizes: torch.Tensor | None = None, dim: int = 0):
    size = tensor.size(dim)
    world_size = get_gp_world_size()
    if size % world_size == 0:
        return tensor, sizes
    trim_size = size - size % world_size
    if dim == 0:
        tensor = tensor[:trim_size]
    elif dim == 1:
        tensor = tensor[:, :trim_size]
    else:
        raise ValueError
    if sizes is not None:
        sizes[-1] = sizes[-1] - size % world_size
    return tensor, sizes


def _tensor_to_split_partitions(tensor: torch.Tensor, dim: int = -1):
    group = get_gp_group()
    num_parts = dist.get_world_size(group=group)
    return [len(part) for part in np.array_split(np.zeros(tensor.size(dim)), num_parts)]


def _split_tensor(
    tensor: torch.Tensor,
    dim: int = -1,
    contiguous_chunks: bool = False,
):
    tensor_list = torch.split(tensor, _tensor_to_split_partitions(tensor, dim), dim=dim)
    if contiguous_chunks:
        return tuple(chunk.contiguous() for chunk in tensor_list)
    return tensor_list


def _reduce(ctx: Any, input: torch.Tensor) -> torch.Tensor:
    group = get_gp_group()
    if ctx:
        ctx.mark_dirty(input)
    dist.all_reduce(input, group=group)
    return input


def _split(input: torch.Tensor, dim: int = -1) -> torch.Tensor:
    rank = get_gp_rank()
    input_list = _split_tensor(input, dim=dim)
    return input_list[rank].clone().contiguous()


def _gather(input: torch.Tensor, dim: int = -1) -> torch.Tensor:
    group = get_gp_group()
    rank = get_gp_rank()
    world_size = dist.get_world_size(group=group)
    if world_size == 1:
        return input
    tensor_list = [torch.empty_like(input) for _ in range(world_size)]
    tensor_list[rank] = input
    dist.all_gather(tensor_list, input, group=group)
    return torch.cat(tensor_list, dim=dim).contiguous()


def _gather_with_padding(input: torch.Tensor, dim: int = -1) -> torch.Tensor:
    group = get_gp_group()
    rank = get_gp_rank()
    world_size = dist.get_world_size(group=group)
    if world_size == 1:
        return input

    # Gather sizes
    size_list = [
        torch.empty(1, device=input.device, dtype=torch.long) for _ in range(world_size)
    ]
    size = torch.tensor([input.size(dim)], device=input.device, dtype=torch.long)
    size_list[rank] = size
    dist.all_gather(size_list, size, group=group)

    # Gather the inputs
    max_size = int(max([size.item() for size in size_list]))
    input = pad_tensor(input, dim, max_size)
    shape = list(input.shape)
    shape[dim] = max_size
    tensor_list = [
        torch.empty(shape, device=input.device, dtype=input.dtype)
        for _ in range(world_size)
    ]

    dist.all_gather(tensor_list, input, group=group)
    tensor_list[rank] = input  # pop back in our local copy (requires grad)

    # Trim and cat
    return torch.cat(
        [tensor.narrow(dim, 0, size) for tensor, size in zip(tensor_list, size_list)],
        dim=dim,
    ).contiguous()


class CopyToModelParallelRegion(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor) -> torch.Tensor:
        return input

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:
        return _reduce(None, grad_output)


class ReduceFromModelParallelRegion(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor) -> torch.Tensor:
        # return _reduce(ctx, input) # this operates in place
        return all_reduce(input, group=get_gp_group())  # this operats out of place

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:
        return grad_output


# this returns the values in place
class ScatterToModelParallelRegion(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor, dim: int = -1) -> torch.Tensor:
        result = _split(input, dim)
        ctx.save_for_backward(torch.tensor(dim))
        return result

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        (dim,) = ctx.saved_tensors
        return _gather_with_padding(grad_output.clone(), dim.item()), None


class GatherFromModelParallelRegion(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor, dim: int = -1) -> torch.Tensor:
        ctx.save_for_backward(torch.tensor(dim))
        return _gather_with_padding(input, dim)

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        (dim,) = ctx.saved_tensors
        result = _split(grad_output, dim.item())
        return result, None


class GatherFromModelParallelRegionSumGrad(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor, dim: int = -1) -> torch.Tensor:
        ctx.save_for_backward(torch.tensor(dim))
        return _gather_with_padding(input, dim)

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        (dim,) = ctx.saved_tensors
        group = get_gp_group()
        # use dist internal # does not work
        # reduced_grad_output = grad_output.clone()
        # dist.all_reduce(
        #    reduced_grad_output, group=group
        # )  # This is an inplace operation
        # grad_output = reduced_grad_output

        # use functional version instead
        grad_output = all_reduce(grad_output, group=group)

        result = _split(grad_output, dim.item())
        return result, None


# Leave forward untouched but upscale the gradient by a factor of gp_group_size
# DDP reduces a mean across the loss, if we have gp_group_size=2 and 6 ranks
# that means we do (a_1+a_2+a_3+b_1+b_2+b_3)/6 in ddp mean. This gets us the
# correct loss but the grad is wrong by a factor of gp_group_size
# dL/d_a1 = 1/6 but it should be dL/da = 1/2 (for the equivalanet non GP run
# with 2 ranks)
# we coud perform an extra round of all_reduce, but this would increase
# communication overhead, instead we can just upscsale the gradient only and
# avoid over head communication
class ScaleBackwardGrad(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor) -> torch.Tensor:
        return input

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        return dist.get_world_size(get_gp_group()) * grad_output


def copy_to_model_parallel_region(input: torch.Tensor) -> torch.Tensor:
    assert initialized(), "Cannot use graph parallel with initializing gp group, must call setup_gp from gp_utils.py!"
    return CopyToModelParallelRegion.apply(input)


def reduce_from_model_parallel_region(input: torch.Tensor) -> torch.Tensor:
    assert initialized(), "Cannot use graph parallel with initializing gp group, must call setup_gp from gp_utils.py!"
    return ReduceFromModelParallelRegion.apply(input)


def scatter_to_model_parallel_region(
    input: torch.Tensor, dim: int = -1
) -> torch.Tensor:
    assert initialized(), "Cannot use graph parallel with initializing gp group, must call setup_gp from gp_utils.py!"
    return ScatterToModelParallelRegion.apply(input, dim)


def gather_from_model_parallel_region(
    input: torch.Tensor, dim: int = -1
) -> torch.Tensor:
    assert initialized(), "Cannot use graph parallel with initializing gp group, must call setup_gp from gp_utils.py!"
    return GatherFromModelParallelRegion.apply(input, dim)


def gather_from_model_parallel_region_sum_grad(
    input: torch.Tensor, dim: int = -1
) -> torch.Tensor:
    assert initialized(), "Cannot use graph parallel with initializing gp group, must call setup_gp from gp_utils.py!"
    return GatherFromModelParallelRegionSumGrad.apply(input, dim)


def scale_backward_grad(input: torch.Tensor) -> torch.Tensor:
    assert initialized(), "Cannot use graph parallel with initializing gp group, must call setup_gp from gp_utils.py!"
    return ScaleBackwardGrad.apply(input)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
from abc import ABCMeta

import hydra
import numpy as np
import omegaconf
import torch
from tqdm import tqdm

from fairchem.core.components.runner import Runner


def count_pairs(tensor, max_value=100):
    # Ensure the tensor is of type long for indexing
    tensor = tensor.long()
    # Flatten the pairs into a single index
    indices = (tensor[:, 0] - 1) * max_value + (tensor[:, 1] - 1)
    # Count occurrences of each index
    counts = torch.bincount(indices, minlength=max_value * max_value)
    count_matrix = counts.view(max_value, max_value)

    return count_matrix


class PairwiseCountRunner(Runner, metaclass=ABCMeta):
    """Perform a single point calculation of several structures/molecules.

    This class handles the single point calculation of atomic structures using a specified calculator,
    processes the input data in chunks, and saves the results.
    """

    def __init__(
        self,
        dataset_cfg="/checkpoint/ocp/shared/pairwise_data/preview_config.yaml",
        ds_name="omat",
        radius=3.5,
        portion=0.01,
    ):
        """
        Args:
            dataset_cfg: Path to the dataset configuration file
            ds_name: Name of the dataset to process
            radius: Radius for the pairwise calculation
            portion: Portion of the dataset to process (default is 100)

        """
        self.dataset_cfg = dataset_cfg
        self.ds_name = ds_name
        self.radius = radius
        self.portion = portion

    def run(self):
        os.makedirs(
            f"/checkpoint/ocp/shared/pairwise_data/uma-preview-r{self.radius}_p{self.portion}",
            exist_ok=True,
        )

        job_num = self.job_config.metadata.array_job_num
        num_jobs = self.job_config.scheduler.num_array_jobs

        # canonical config of a training run. here I took uma_sm_direct (preview)
        cfg = omegaconf.OmegaConf.load(self.dataset_cfg)
        dataset_cfg = cfg["runner"]["train_dataloader"]["dataset"]
        dataset_names = sorted(dataset_cfg["dataset_configs"].keys())
        for ds in dataset_names:
            dataset_cfg["dataset_configs"][ds]["a2g_args"]["radius"] = self.radius
            dataset_cfg["dataset_configs"][ds]["a2g_args"]["max_neigh"] = 300

        concat_uma_dataset = hydra.utils.instantiate(dataset_cfg)

        ds_idx = dataset_names.index(self.ds_name)
        dataset = concat_uma_dataset.datasets[ds_idx]

        count_mtx = np.zeros([100, 100])
        downsample = int(1 / self.portion)
        chunk_indices = np.array_split(range(len(dataset)), num_jobs)[job_num][
            ::downsample
        ]
        for idx in tqdm(chunk_indices):
            data = dataset[idx]
            pairs = data.atomic_numbers[data.edge_index].T
            count_mtx += count_pairs(pairs).float().numpy()

        np.save(
            f"/checkpoint/ocp/shared/pairwise_data/uma-preview-r{self.radius}_p{self.portion}/{self.ds_name}_{num_jobs}_{job_num}.npy",
            count_mtx,
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from .calculate_runner import CalculateRunner
from .elasticity_runner import ElasticityRunner
from .relaxation_runner import RelaxationRunner

__all__ = ["CalculateRunner", "ElasticityRunner", "RelaxationRunner"]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, ClassVar

import numpy as np
import pandas as pd
from ase.optimize import LBFGS
from tqdm import tqdm

from fairchem.core.components.calculate import CalculateRunner
from fairchem.core.components.calculate.recipes.adsorption import (
    adsorb_atoms,
)

if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator
    from ase.optimize import Optimizer

    from fairchem.core.datasets import AseDBDataset


class AdsorptionRunner(CalculateRunner):
    """
    Relax an adsorbate+surface configuration to compute the adsorption energy.
    The option to also relax a clean surface is also provided.

    This class handles the relaxation of atomic structures using a specified calculator,
    processes the input data in chunks, and saves the results.

    Input data is an AseDBDataset where each atoms object is organized as
    follows:
        atoms: adsorbate+surface configuration
        atoms.info = {
            gas_ref: float,
            dft_relaxed_adslab_energy: float,
            dft_relaxed_slab_energy: float,
            initial_slab_atoms: ase.Atoms, # Required if relax_surface=True
        }
    """

    result_glob_pattern: ClassVar[str] = "adsorption_*-*.json.gz"

    def __init__(
        self,
        calculator: Calculator,
        input_data: AseDBDataset,
        save_relaxed_atoms: bool = True,
        relax_surface: bool = False,
        optimizer_cls: type[Optimizer] = LBFGS,
        fmax: float = 0.05,
        steps: int = 300,
    ):
        """
        Initialize the AdsorptionRunner.

        Args:
            calculator: ASE calculator to use for energy and force calculations
            input_data: Dataset containing atomic structures to process
            save_relaxed_atoms (bool): Whether to save the relaxed structures in the results
            relax_surface (bool): Whether to relax the bare surface
            optimizer_cls (Optimizer): ASE optimizer class to use
            fmax (float): force convergence threshold
            steps (int): max number of relaxation steps
        """
        self._save_relaxed_atoms = save_relaxed_atoms
        self.fmax = fmax
        self.steps = steps
        self.relax_surface = relax_surface
        self.optimizer_cls = optimizer_cls
        super().__init__(calculator=calculator, input_data=input_data)

    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        """Perform relaxation calculations on a subset of structures.

        Splits the input data into chunks and processes the chunk corresponding to job_num.

        Args:
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.

        Returns:
            list[dict[str, Any]] - List of dictionaries containing calculation results
        """
        all_results = []
        chunk_indices = np.array_split(range(len(self.input_data)), num_jobs)[job_num]
        for i in tqdm(chunk_indices, desc="Running relaxations"):
            atoms = self.input_data.get_atoms(i)
            results = adsorb_atoms(
                atoms,
                self.calculator,
                optimizer_cls=self.optimizer_cls,
                steps=self.steps,
                fmax=self.fmax,
                relax_surface=self.relax_surface,
                save_relaxed_atoms=self._save_relaxed_atoms,
            )
            all_results.append(results)

        return all_results

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        """Write calculation results to a compressed JSON file.

        Args:
            results: List of dictionaries containing elastic properties
            results_dir: Directory path where results will be saved
            job_num: Index of the current job
            num_jobs: Total number of jobs
        """
        results_df = pd.DataFrame(results)
        results_df.to_json(
            os.path.join(results_dir, f"adsorption_{num_jobs}-{job_num}.json.gz")
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



from __future__ import annotations

from dataclasses import MISSING, dataclass, field, fields, is_dataclass
from typing import Literal


@dataclass
class GlobalConfigs:
    regress_forces: bool
    direct_forces: bool
    hidden_size: int  # divisible by 2 and num_heads
    num_layers: int
    activation: Literal[
        "squared_relu", "gelu", "leaky_relu", "relu", "smelu", "star_relu"
    ] = "gelu"
    regress_stress: bool = False
    use_compile: bool = True
    use_padding: bool = True
    use_fp16_backbone: bool = False
    dataset_list: list = field(default_factory=list)


@dataclass
class MolecularGraphConfigs:
    use_pbc: bool
    max_num_elements: int
    max_atoms: int
    max_batch_size: int
    max_radius: float
    knn_k: int
    knn_soft: bool
    knn_sigmoid_scale: float
    knn_lse_scale: float
    knn_use_low_mem: bool
    knn_pad_size: int
    distance_function: Literal["gaussian", "sigmoid", "linearsigmoid", "silu"] = (
        "gaussian"
    )
    use_envelope: bool = True


@dataclass
class GraphNeuralNetworksConfigs:
    atten_name: Literal[
        "math",
        "memory_efficient",
        "flash",
    ]
    atten_num_heads: int
    atom_embedding_size: int = 128
    node_direction_embedding_size: int = 64
    node_direction_expansion_size: int = 10
    edge_distance_expansion_size: int = 600
    edge_distance_embedding_size: int = 512
    readout_hidden_layer_multiplier: int = 2
    output_hidden_layer_multiplier: int = 2
    ffn_hidden_layer_multiplier: int = 2
    use_angle_embedding: Literal["scalar", "bias", "none"] = "none"
    angle_expansion_size: int = 10
    angle_embedding_size: int = 8
    use_graph_attention: bool = False
    use_message_gate: bool = False
    use_global_readout: bool = False
    use_frequency_embedding: bool = True
    freequency_list: list = field(default_factory=lambda: [20, 10, 4, 10, 20])
    energy_reduce: Literal["sum", "mean"] = "sum"


@dataclass
class RegularizationConfigs:
    normalization: Literal["layernorm", "rmsnorm", "skip"] = "rmsnorm"
    mlp_dropout: float = 0.0
    atten_dropout: float = 0.0
    stochastic_depth_prob: float = 0.0
    node_ffn_dropout: float = 0.0
    edge_ffn_dropout: float = 0.0
    scalar_output_dropout: float = 0.0
    vector_output_dropout: float = 0.0


@dataclass
class EScAIPConfigs:
    global_cfg: GlobalConfigs
    molecular_graph_cfg: MolecularGraphConfigs
    gnn_cfg: GraphNeuralNetworksConfigs
    reg_cfg: RegularizationConfigs


def resolve_type_hint(cls, field):
    """Resolves forward reference type hints from string to actual class objects."""
    if isinstance(field.type, str):
        resolved_type = getattr(cls, field.type, None)
        if resolved_type is None:
            resolved_type = globals().get(field.type, None)  # Try global scope
        if resolved_type is None:
            return field.type  # Fallback to string if not found
        return resolved_type
    return field.type


def init_configs(cls, kwargs):
    """
    Initialize a dataclass with the given kwargs.
    """
    init_kwargs = {}
    for _field in fields(cls):
        field_name = _field.name
        field_type = resolve_type_hint(cls, _field)  # Resolve type

        if is_dataclass(field_type):  # Handle nested dataclass
            init_kwargs[_field.name] = init_configs(field_type, kwargs)
        elif field_name in kwargs:  # Direct assignment
            init_kwargs[field_name] = kwargs[field_name]
        elif _field.default is not MISSING:  # Assign default if available
            init_kwargs[field_name] = _field.default
        elif _field.default_factory is not MISSING:  # Handle default_factory
            init_kwargs[field_name] = _field.default_factory()
        else:
            raise ValueError(
                f"Missing required configuration parameter: '{field_name}' in '{cls.__name__}'"
            )

    return cls(**init_kwargs)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import multiprocessing
import os
import pdb
import sys
from dataclasses import dataclass
from datetime import timedelta
from typing import Any

import torch
import torch.distributed as dist
from torch.distributed.elastic.utils.distributed import get_free_port

from fairchem.core.common.gp_utils import setup_gp


class ForkedPdb(pdb.Pdb):
    """A Pdb subclass that may be used from a forked multiprocessing child
    https://stackoverflow.com/questions/4716533/how-to-attach-debugger-to-a-python-subproccess/23654936#23654936

    example usage to debug a torch distributed run on rank 0:
    if torch.distributed.get_rank() == 0:
        from fairchem.core.common.test_utils import ForkedPdb
        ForkedPdb().set_trace()
    """

    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            with open("/dev/stdin") as f:
                sys.stdin = f
                pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin


@dataclass
class PGConfig:
    backend: str
    world_size: int
    gp_group_size: int = 1
    port: str = "12345"
    use_gp: bool = True


def init_env_rank_and_launch_test(
    rank: int,
    pg_setup_params: PGConfig,
    mp_output_dict: dict[int, object],
    test_method: callable,
    args: list[object],
    kwargs: dict[str, object],
) -> None:
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = pg_setup_params.port
    os.environ["WORLD_SIZE"] = str(pg_setup_params.world_size)
    os.environ["LOCAL_RANK"] = str(rank)
    os.environ["RANK"] = str(rank)
    mp_output_dict[rank] = test_method(*args, **kwargs)  # pyre-fixme


def init_pg_and_rank_and_launch_test(
    rank: int,
    pg_setup_params: PGConfig,
    mp_output_dict: dict[int, object],
    test_method: callable,
    args: list[object],
    kwargs: dict[str, object],
) -> None:
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = pg_setup_params.port
    os.environ["WORLD_SIZE"] = str(pg_setup_params.world_size)
    os.environ["LOCAL_RANK"] = str(rank)
    # setup default process group
    dist.init_process_group(
        rank=rank,
        world_size=pg_setup_params.world_size,
        backend=pg_setup_params.backend,
        timeout=timedelta(seconds=10),  # setting up timeout for distributed collectives
    )
    # setup gp
    if pg_setup_params.use_gp:
        config = {
            "gp_gpus": pg_setup_params.gp_group_size,
            "distributed_backend": pg_setup_params.backend,
        }
        setup_gp(config)
    mp_output_dict[rank] = test_method(*args, **kwargs)  # pyre-fixme


def spawn_multi_process(
    config: PGConfig,
    test_method: callable,
    init_and_launch: callable,
    *test_method_args: Any,
    **test_method_kwargs: Any,
) -> list[Any]:
    """
    Spawn single node, multi-rank function.
    Uses localhost and free port to communicate.

    Args:
        world_size: number of processes
        backend: backend to use. for example, "nccl", "gloo", etc
        test_method: callable to spawn. first 3 arguments are rank, world_size and mp output dict
        test_method_args: args for the test method
        test_method_kwargs: kwargs for the test method

    Returns:
        A list, l, where l[i] is the return value of test_method on rank i
    """
    manager = multiprocessing.Manager()
    mp_output_dict = manager.dict()

    port = str(get_free_port())
    config.port = port
    torch.multiprocessing.spawn(
        # torch.multiprocessing.spawn sends rank as the first param
        # https://pytorch.org/docs/stable/multiprocessing.html#torch.multiprocessing.spawn
        init_and_launch,
        args=(
            config,
            mp_output_dict,
            test_method,
            test_method_args,
            test_method_kwargs,
        ),
        nprocs=config.world_size,
    )

    return [mp_output_dict[i] for i in range(config.world_size)]


def init_local_distributed_process_group(backend="nccl"):
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = str(get_free_port())
    dist.init_process_group(
        rank=0,
        world_size=1,
        backend=backend,
        timeout=timedelta(seconds=10),  # setting up timeout for distributed collectives
    )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
import shutil

CACHE_DIR = os.environ.get(
    "FAIRCHEM_CACHE_DIR", os.path.join(os.path.expanduser("~"), ".cache/fairchem")
)
os.makedirs(CACHE_DIR, exist_ok=True)


def clear_cache():
    try:
        shutil.rmtree(CACHE_DIR)
    except FileNotFoundError:
        print(f"No FAIRChem cache directory found at {CACHE_DIR}")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import torch

if TYPE_CHECKING:
    from fairchem.core.datasets.atomic_data import AtomicData
    from fairchem.core.models.escaip.configs import (
        GlobalConfigs,
        GraphNeuralNetworksConfigs,
        MolecularGraphConfigs,
    )
from fairchem.core.models.escaip.custom_types import GraphAttentionData
from fairchem.core.models.escaip.utils.graph_utils import (
    get_attn_mask,
    get_attn_mask_env,
    get_compact_frequency_vectors,
    get_node_direction_expansion_neighbor,
    pad_batch,
)
from fairchem.core.models.escaip.utils.radius_graph import (
    biknn_radius_graph,
    envelope_fn,
    safe_norm,
    safe_normalize,
)
from fairchem.core.models.escaip.utils.smearing import (
    GaussianSmearing,
    LinearSigmoidSmearing,
    SigmoidSmearing,
    SiLUSmearing,
)


def data_preprocess_radius_graph(
    data: AtomicData,
    # generate_graph_fn: callable,
    global_cfg: GlobalConfigs,
    gnn_cfg: GraphNeuralNetworksConfigs,
    molecular_graph_cfg: MolecularGraphConfigs,
) -> GraphAttentionData:
    # atomic numbers
    atomic_numbers = data.atomic_numbers.long()

    # edge distance expansion
    expansion_func = {
        "gaussian": GaussianSmearing,
        "sigmoid": SigmoidSmearing,
        "linear_sigmoid": LinearSigmoidSmearing,
        "silu": SiLUSmearing,
    }[molecular_graph_cfg.distance_function]

    edge_distance_expansion_func = expansion_func(
        0.0,
        molecular_graph_cfg.max_radius,
        gnn_cfg.edge_distance_expansion_size,
        basis_width_scalar=2.0,
    ).to(data.pos.device)

    # generate graph
    (
        disp,  # (num_nodes, num_neighbors, 3)
        src_env,
        _,
        _,
        _,
        neighbor_index,  # (2, num_nodes, num_neighbors)
    ) = biknn_radius_graph(  # type: ignore
        data,
        molecular_graph_cfg.max_radius,  # type: ignore[arg-type]
        molecular_graph_cfg.knn_k,
        molecular_graph_cfg.knn_soft,
        molecular_graph_cfg.knn_sigmoid_scale,
        molecular_graph_cfg.knn_lse_scale,
        molecular_graph_cfg.knn_use_low_mem,
        molecular_graph_cfg.knn_pad_size,
        molecular_graph_cfg.use_pbc,
        data.pos.device,
    )

    num_nodes, num_neighbors, _ = disp.shape
    edge_direction = safe_normalize(disp)  # (num_nodes, num_neighbors, 3)
    edge_distance = safe_norm(disp)  # (num_nodes, num_neighbors)
    neighbor_mask = src_env != torch.inf  # (num_nodes, num_neighbors)
    src_mask = envelope_fn(src_env, molecular_graph_cfg.use_envelope)

    # edge distance expansion (ref: scn)
    # (num_nodes, num_neighbors, edge_distance_expansion_size)
    edge_distance_expansion = edge_distance_expansion_func(edge_distance.flatten())
    edge_distance_expansion = edge_distance_expansion.view(
        num_nodes, num_neighbors, gnn_cfg.edge_distance_expansion_size
    )

    # node direction expansion (num_nodes, num_neighbors, lmax + 1)
    node_direction_expansion = get_node_direction_expansion_neighbor(
        direction_vec=edge_direction,
        neighbor_mask=neighbor_mask,
        lmax=gnn_cfg.node_direction_expansion_size - 1,
    )

    neighbor_list = neighbor_index[1]

    # pad batch
    if global_cfg.use_padding:
        (
            atomic_numbers,
            node_direction_expansion,
            edge_distance_expansion,
            edge_direction,
            neighbor_list,
            neighbor_mask,
            src_mask,
            node_batch,
            node_padding_mask,
            graph_padding_mask,
        ) = pad_batch(
            max_atoms=molecular_graph_cfg.max_atoms,
            max_batch_size=molecular_graph_cfg.max_batch_size,
            atomic_numbers=atomic_numbers,
            node_direction_expansion=node_direction_expansion,
            edge_distance_expansion=edge_distance_expansion,
            edge_direction=edge_direction,
            neighbor_list=neighbor_list,
            neighbor_mask=neighbor_mask,
            src_mask=src_mask,
            node_batch=data.batch,
            num_graphs=data.num_graphs,
        )
    else:
        node_padding_mask = torch.ones_like(atomic_numbers, dtype=torch.bool)
        graph_padding_mask = torch.ones(
            data.num_graphs, dtype=torch.bool, device=data.batch.device
        )
        node_batch = data.batch

    # patch singleton atom (TODO: check if this is needed)
    # if global_cfg.use_padding:
    #     edge_direction, neighbor_list, neighbor_mask = patch_singleton_atom(
    #         edge_direction, neighbor_list, neighbor_mask
    #     )

    # get attention mask
    attn_mask = get_attn_mask_env(src_mask, gnn_cfg.atten_num_heads)  # type: ignore
    if gnn_cfg.use_angle_embedding == "none":
        angle_embedding = None
    else:
        attn_mask, angle_embedding = get_attn_mask(
            edge_direction=edge_direction,
            neighbor_mask=neighbor_mask,
            num_heads=gnn_cfg.atten_num_heads,
            lmax=gnn_cfg.node_direction_expansion_size,
            use_angle_embedding=gnn_cfg.use_angle_embedding,
        )

    # get frequency vectors
    if gnn_cfg.use_frequency_embedding:
        # Create repeating dimensions tensor for assertion checks
        repeating_dimensions = torch.tensor(
            gnn_cfg.freequency_list, dtype=torch.long, device=data.pos.device
        )

        # Add assertions to validate repeating_dimensions
        # Check that values sum to head_dim
        head_dim = global_cfg.hidden_size // gnn_cfg.atten_num_heads
        sum_repeats = torch.sum(repeating_dimensions).item()
        assert sum_repeats == head_dim, (
            f"Sum of freequency_list must equal head_dim ({head_dim}), "
            f"but got sum={sum_repeats}. Please adjust freequency_list."
        )

        # Use the Python list directly for better torch.compile compatibility
        frequency_vectors = get_compact_frequency_vectors(
            edge_direction=edge_direction,
            lmax=len(gnn_cfg.freequency_list) - 1,
            repeating_dimensions=gnn_cfg.freequency_list,  # Pass list instead of tensor
        )
    else:
        frequency_vectors = None

    if gnn_cfg.atten_name in ["memory_efficient", "flash", "math"]:
        if (
            gnn_cfg.atten_name in ["memory_efficient", "flash"]
            and not global_cfg.direct_forces
        ):
            logging.warning(
                "Fallback to math attention for gradient based force prediction"
            )
            gnn_cfg.atten_name = "math"
        torch.backends.cuda.enable_flash_sdp(gnn_cfg.atten_name == "flash")
        torch.backends.cuda.enable_mem_efficient_sdp(
            gnn_cfg.atten_name == "memory_efficient"
        )
        # enable math attention for fallbacks
        torch.backends.cuda.enable_math_sdp(True)
    else:
        raise NotImplementedError(
            f"Attention name {gnn_cfg.atten_name} not implemented"
        )

    # construct input data
    x = GraphAttentionData(
        atomic_numbers=atomic_numbers,
        node_direction_expansion=node_direction_expansion,
        edge_distance_expansion=edge_distance_expansion,
        edge_direction=edge_direction,
        attn_mask=attn_mask,
        angle_embedding=angle_embedding,
        frequency_vectors=frequency_vectors,
        neighbor_list=neighbor_list,
        neighbor_mask=neighbor_mask,
        node_batch=node_batch,
        node_padding_mask=node_padding_mask,
        graph_padding_mask=graph_padding_mask,
    )
    return x



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import glob
import os

import numpy as np

SPLITS = {
    "OC20": ["id", "ood_ads", "ood_cat", "ood_both"],
    "OC22": ["id", "ood"],
}


def write_is2re_relaxations(args) -> None:
    import ase.io
    from tqdm import tqdm

    submission_file = {}

    if not args.hybrid:
        for split in SPLITS[args.dataset]:
            ids = []
            energies = []
            systems = glob.glob(os.path.join(vars(args)[split], "*.traj"))
            for system in tqdm(systems):
                sid, _ = os.path.splitext(os.path.basename(system))
                ids.append(str(sid))
                # Read the last frame in the ML trajectory. Modify "-1" if you wish to modify which frame to use.
                traj = ase.io.read(system, "-1")
                energies.append(traj.get_potential_energy())

            submission_file[f"{split}_ids"] = np.array(ids)
            submission_file[f"{split}_energy"] = np.array(energies)

    else:
        for split in SPLITS[args.dataset]:
            preds = np.load(vars(args)[split])
            ids = []
            energies = []
            for sid, energy in zip(preds["ids"], preds["energy"]):
                ids.append(sid.split("_")[0])
                energies.append(energy)

            submission_file[f"{split}_ids"] = np.array(ids)
            submission_file[f"{split}_energy"] = np.array(energies)

    np.savez_compressed(args.out_path, **submission_file)


def write_predictions(args) -> None:
    if args.is2re_relaxations:
        write_is2re_relaxations(args)
    else:
        submission_file = {}

        for split in SPLITS[args.dataset]:
            res = np.load(vars(args)[split], allow_pickle=True)
            contents = res.files
            for i in contents:
                key = f"{split}_{i}"
                submission_file[key] = res[i]

        np.savez_compressed(args.out_path, **submission_file)


def main(args: argparse.Namespace) -> None:
    for split in SPLITS[args.dataset]:
        assert vars(args).get(split), f"Missing {split} split for {args.dataset}"

    if not args.out_path.endswith(".npz"):
        args.out_path = args.out_path + ".npz"

    write_predictions(args)
    print(f"Results saved to {args.out_path} successfully.")


if __name__ == "__main__":
    """
    Create a submission file for evalAI. Ensure that for the task you are
    submitting for you have generated results files on each of the splits:
        OC20: id, ood_ads, ood_cat, ood_both
        OC22: id, ood

    Results file can be obtained as follows for the various tasks:

    S2EF: config["mode"] = "predict"
    IS2RE: config["mode"] = "predict"
    IS2RS: config["mode"] = "run-relaxations" and config["task"]["write_pos"] = True

    Use this script to join the results files (4 for OC20, 2 for OC22) in the format evalAI expects
    submissions.

    If writing IS2RE predictions from relaxations, paths must be directories
    containg trajectory files. Additionally, --is2re-relaxations must be
    provided as a command line argument.

    If writing IS2RE predictions from hybrid relaxations (force only model +
    energy only model), paths must be the .npz S2EF prediction files.
    Additionally, --is2re-relaxations and --hybrid must be provided as a
    command line argument.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--id", help="Path to ID results. Required for OC20 and OC22.")
    parser.add_argument(
        "--ood-ads", help="Path to OOD-Ads results. Required only for OC20."
    )
    parser.add_argument(
        "--ood-cat", help="Path to OOD-Cat results. Required only for OC20."
    )
    parser.add_argument(
        "--ood-both", help="Path to OOD-Both results. Required only for OC20."
    )
    parser.add_argument(
        "--ood", help="Path to OOD OC22 results. Required only for OC22."
    )
    parser.add_argument("--out-path", help="Path to write predictions to.")
    parser.add_argument(
        "--is2re-relaxations",
        action="store_true",
        help="Write IS2RE results from trajectories. Paths specified correspond to directories containing .traj files.",
    )
    parser.add_argument(
        "--hybrid",
        action="store_true",
        help="Write IS2RE results from S2EF prediction files. Paths specified correspond to S2EF NPZ files.",
    )
    parser.add_argument(
        "--dataset",
        type=str,
        default="OC20",
        choices=["OC20", "OC22"],
        help="Which dataset to write a prediction file for, OC20 or OC22.",
    )

    args: argparse.Namespace = parser.parse_args()
    main(args)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
from functools import partial
from typing import TYPE_CHECKING, Any, Literal, Optional

from ase.atoms import Atoms
from fairchem.data.oc.core.adsorbate import Adsorbate
from fairchem.data.oc.core.multi_adsorbate_slab_config import (
    MultipleAdsorbateSlabConfig,
)
from fairchem.data.oc.core.slab import Slab
from fairchem.data.oc.utils import DetectTrajAnomaly

if TYPE_CHECKING:
    from collections.abc import Callable

    from ase.optimize import Optimizer


def relax_job(initial_atoms, calc, optimizer_cls, fmax, steps):
    atoms = initial_atoms.copy()
    atoms.calc = calc
    dyn = optimizer_cls(atoms)
    dyn.run(fmax=fmax, steps=steps)
    energy = atoms.get_potential_energy()
    forces = atoms.get_forces()
    atoms.calc = None
    result = {
        "input_atoms": {"atoms": initial_atoms},
        "atoms": atoms,
        "results": {
            "energy": energy,
            "forces": forces,
        },
    }
    return result


def adsorb_ml_pipeline(
    slab: Slab,
    adsorbates_kwargs: dict[str, Any],
    multiple_adsorbate_slab_config_kwargs: dict[str, Any],
    ml_slab_adslab_relax_job: Callable[..., Any],
    reference_ml_energies: bool = True,
    atomic_reference_energies: Optional[dict] = None,
    relaxed_slab_atoms: Atoms = None,
    place_on_relaxed_slab: bool = False,
):
    """
    Run a machine learning-based pipeline for adsorbate-slab systems.

    1. Relax slab using ML
    2. Generate trial adsorbate-slab configurations for the relaxed slab
    3. Relax adsorbate-slab configurations using ML
    4. Validate slab and adsorbate-slab configurations (check for anomalies like dissociations))
    5. Reference the energies to gas phase if needed (eg using a total energy ML model)

    Parameters
    ----------
    slab : Slab
        The slab structure to which adsorbates will be added.
    adsorbates_kwargs : dict[str,Any]
        Keyword arguments for generating adsorbate configurations.
    multiple_adsorbate_slab_config_kwargs : dict[str, Any]
        Keyword arguments for generating multiple adsorbate-slab configurations.
    ml_slab_adslab_relax_job : Job
        Job for relaxing slab and adsorbate-slab configurations using ML.
    reference_ml_energies: bool, optional
        Whether to reference ML energies to gas phase, by default False.
    atomic_reference_energies : AtomicReferenceEnergies, optional
        Atomic reference energies for referencing, by default None.
    relaxed_slab_atoms: ase.Atoms, optional
        DFT Relaxed slab atoms for anomaly detection for adsorption energy models, by default None.
    place_on_relaxed_slab: bool, optional
        Whether to place adsorbates on the relaxed slab or initial unrelaxed slab, by default False.

    Returns
    -------
    dict
        Dictionary containing the slab, ML-relaxed adsorbate-slab configurations,
        detected anomalies.
    """

    # only run slab relaxation if total energy model or placing on relaxed slab
    if place_on_relaxed_slab or reference_ml_energies:
        ml_relaxed_slab_result = ml_slab_adslab_relax_job(slab.atoms)

    unrelaxed_adslab_configurations = ocp_adslab_generator(
        ml_relaxed_slab_result["atoms"] if place_on_relaxed_slab else slab.atoms,
        adsorbates_kwargs,
        multiple_adsorbate_slab_config_kwargs,
    )

    ml_relaxed_configurations = [
        ml_slab_adslab_relax_job(adslab_configuration)
        for adslab_configuration in unrelaxed_adslab_configurations
    ]

    if reference_ml_energies:
        assert (
            atomic_reference_energies is not None
        ), "Missing atomic reference energies"

        ml_relaxed_configurations = reference_adslab_energies(
            ml_relaxed_configurations,
            ml_relaxed_slab_result,
            atomic_energies=atomic_reference_energies,
        )

    adslab_anomalies_list = [
        detect_anomaly(
            relaxed_result["input_atoms"]["atoms"],
            relaxed_result["atoms"],
            ml_relaxed_slab_result["atoms"]
            if reference_ml_energies
            else relaxed_slab_atoms,
        )
        for relaxed_result in ml_relaxed_configurations
    ]

    top_candidates = filter_sort_select_adslabs(
        adslab_results=ml_relaxed_configurations,
        adslab_anomalies_list=adslab_anomalies_list,
    )

    return {
        "slab": slab.get_metadata_dict(),
        "adslabs": top_candidates,
        "adslab_anomalies": adslab_anomalies_list,
    }


def ocp_adslab_generator(
    slab: Slab | Atoms,
    adsorbates_kwargs: list[dict[str, Any]] | None = None,
    multiple_adsorbate_slab_config_kwargs: dict[str, Any] | None = None,
) -> list[Atoms]:
    """
    Generate adsorbate-slab configurations.

    Parameters
    ----------
    slab : Slab | Atoms
        The slab structure.
    adsorbates_kwargs : list[dict[str,Any]], optional
        List of keyword arguments for generating adsorbates, by default None.
    multiple_adsorbate_slab_config_kwargs : dict[str,Any], optional
        Keyword arguments for generating multiple adsorbate-slab configurations, by default None.

    Returns
    -------
    list[Atoms]
        List of generated adsorbate-slab configurations.
    """
    adsorbates = [
        Adsorbate(**adsorbate_kwargs) for adsorbate_kwargs in adsorbates_kwargs
    ]

    if isinstance(slab, Atoms):
        slab = Slab(slab_atoms=slab)

    if multiple_adsorbate_slab_config_kwargs is None:
        multiple_adsorbate_slab_config_kwargs = {}

    adslabs = MultipleAdsorbateSlabConfig(
        copy.deepcopy(slab), adsorbates, **multiple_adsorbate_slab_config_kwargs
    )

    atoms_list = adslabs.atoms_list
    for atoms in atoms_list:
        atoms.pbc = True

    return adslabs.atoms_list


def reference_adslab_energies(
    adslab_results: list[dict],
    slab_result: dict,
    atomic_energies: dict,
) -> list[dict]:
    """
    Reference adsorbate-slab energies to atomic and slab energies.

    Parameters
    ----------
    adslab_results : list[dict[str, Any]]
        List of adsorbate-slab results.
    slab_result : dict
        Result of the slab calculation.
    atomic_energies : AtomicReferenceEnergies | None
        Dictionary of atomic energies.

    Returns
    -------
    list[dict[str, Any]]
        List of adsorbate-slab results with referenced energies.
    """
    slab_energy = slab_result["results"]["energy"]

    for adslab_result in adslab_results:
        adslab_result["results"]["referenced_adsorption_energy"] = {
            "atomic_energies": atomic_energies,
            "slab_energy": slab_energy,
            "adslab_energy": adslab_result["results"]["energy"],
            "gas_reactant_energy": sum(
                [
                    atomic_energies[atom.symbol]
                    for atom in adslab_result["atoms"][
                        adslab_result["atoms"].get_tags() == 2
                    ]  # all adsorbate tagged with tag=2!
                ]
            ),
            "adsorption_energy": adslab_result["results"]["energy"]
            - slab_energy
            - sum(
                [
                    atomic_energies[atom.symbol]
                    for atom in adslab_result["atoms"][
                        adslab_result["atoms"].get_tags() == 2
                    ]  # all adsorbate tagged with tag=2!
                ]
            ),
        }

    return adslab_results


def filter_sort_select_adslabs(
    adslab_results: list[dict], adslab_anomalies_list: list[list[str]]
) -> list[dict]:
    """
    Filter, sort, and select adsorbate-slab configurations based on anomalies and energy.

    Parameters
    ----------
    adslab_results : list[dict]
        List of adsorbate-slab results.
    adslab_anomalies_list : list[list[str]]
        List of detected anomalies for each adsorbate-slab configuration.

    Returns
    -------
    list[dict]
        Sorted list of adsorbate-slab configurations without anomalies.
    """
    for adslab_result, adslab_anomalies in zip(
        adslab_results, adslab_anomalies_list, strict=True
    ):
        adslab_result["results"]["adslab_anomalies"] = adslab_anomalies

    adslabs_no_anomalies = [
        adslab_result
        for adslab_result in adslab_results
        if len(adslab_result["results"]["adslab_anomalies"]) == 0
    ]

    return sorted(adslabs_no_anomalies, key=lambda x: x["results"]["energy"])


def detect_anomaly(
    initial_atoms: Atoms,
    final_atoms: Atoms,
    final_slab_atoms: Atoms,
) -> list[
    Literal[
        "adsorbate_dissociated",
        "adsorbate_desorbed",
        "surface_changed",
        "adsorbate_intercalated",
    ]
]:
    """
    Detect anomalies between initial and final atomic structures.

    Parameters
    ----------
    initial_atoms : Atoms
        Initial atomic structure.
    final_atoms : Atoms
        Final atomic structure.

    Returns
    -------
    list[Literal["adsorbate_dissociated", "adsorbate_desorbed", "surface_changed", "adsorbate_intercalated"]]
        List of detected anomalies.
    """
    atom_tags = initial_atoms.get_tags()

    detector = DetectTrajAnomaly(
        initial_atoms,
        final_atoms,
        atoms_tag=atom_tags,
        final_slab_atoms=final_slab_atoms,
    )
    anomalies = []
    if detector.is_adsorbate_dissociated():
        anomalies.append("adsorbate_dissociated")
    if detector.is_adsorbate_desorbed():
        anomalies.append("adsorbate_desorbed")
    if detector.has_surface_changed():
        anomalies.append("surface_changed")
    if detector.is_adsorbate_intercalated():
        anomalies.append("adsorbate_intercalated")
    return anomalies


# Run the AdsorbML workflow for a given slab and adsorbate
def run_adsorbml(
    slab,
    adsorbate,
    calculator,
    optimizer_cls: Optimizer,
    fmax: float = 0.02,
    steps: int = 300,
    num_placements: int = 100,
    reference_ml_energies: bool = True,
    relaxed_slab_atoms: Atoms = None,
    place_on_relaxed_slab: bool = False,
):
    """
    Run the AdsorbML pipeline for a given slab and adsorbate using a pretrained ML model.
    Parameters
    ----------
    slab : ase.Atoms
        The clean slab structure to which the adsorbate will be added.
    adsorbate : str
        A string identifier for the adsorbate from the database (e.g., '*O').
    reference_ml_energies : bool, optional
        If True, assumes the model is a total energy model and references energies
        to gas phase and bare slab, by default True since the default model is a total energy model.
    num_placements : int, optional
        Number of initial adsorbate placements to generate for relaxation, by default 100.
    fmax: float, default 0.02.
        Relaxation force convergence threshold
    steps: int, default 300
        Max number of relaxation steps
    relaxed_slab_atoms : ase.Atoms, optional
        DFT Relaxed slab atoms for anomaly detection for adsorption energy models, by default None.
    place_on_relaxed_slab : bool, optional
        Whether to place adsorbates on the relaxed slab or initial unrelaxed slab, by default False.
    Returns
    -------
    dict
        Dictionary containing the ML-relaxed slab, adsorbate-slab configurations,
        energies, and validation results (matching the AdsorbMLSchema format).
    """

    # if we are using a total energy model, we need to set the DFT atomic reference energies
    # obtained from the supplementary information of the OC20 paper
    atomic_reference_energies = {
        "H": -3.477,  # eV
        "O": -7.204,  # eV
        "C": -7.282,  # eV
        "N": -8.083,  # eV
    }

    ml_relax_job = partial(
        relax_job, calc=calculator, optimizer_cls=optimizer_cls, fmax=fmax, steps=steps
    )

    outputs = adsorb_ml_pipeline(
        slab=slab,
        adsorbates_kwargs=[{"adsorbate_smiles_from_db": adsorbate}],
        multiple_adsorbate_slab_config_kwargs={"num_configurations": num_placements},
        ml_slab_adslab_relax_job=ml_relax_job,
        reference_ml_energies=reference_ml_energies,
        atomic_reference_energies=atomic_reference_energies,
        relaxed_slab_atoms=relaxed_slab_atoms,
        place_on_relaxed_slab=place_on_relaxed_slab,
    )
    return outputs



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import logging
import os
import random
import tempfile
import uuid
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

import clusterscope
import hydra
import numpy as np
import torch
from omegaconf import OmegaConf
from omegaconf.errors import InterpolationKeyError

from fairchem.core.common import gp_utils

if TYPE_CHECKING:
    from omegaconf import DictConfig

    from fairchem.core.components.reducer import Reducer
    from fairchem.core.components.runner import Runner

from submitit import AutoExecutor
from submitit.core.utils import JobPaths, cloudpickle_dump
from submitit.helpers import Checkpointable, DelayedSubmission
from submitit.slurm.slurm import SlurmJobEnvironment

from fairchem.core.common import distutils
from fairchem.core.common.logger import WandBSingletonLogger
from fairchem.core.common.utils import (
    StrEnum,
    get_commit_hash,
    get_timestamp_uid,
    setup_env_vars,
    setup_logging,
)

# this effects the cli only since the actual job will be run in subprocesses or remoe
logging.basicConfig(level=logging.INFO)


ALLOWED_TOP_LEVEL_KEYS = {"job", "runner", "reducer"}

LOG_DIR_NAME = "logs"
CHECKPOINT_DIR_NAME = "checkpoints"
RESULTS_DIR = "results"
CONFIG_FILE_NAME = "canonical_config.yaml"
PREEMPTION_STATE_DIR_NAME = "preemption_state"


class SchedulerType(StrEnum):
    LOCAL = "local"
    SLURM = "slurm"


class DeviceType(StrEnum):
    CPU = "cpu"
    CUDA = "cuda"


class RunType(StrEnum):
    RUN = "run"
    REDUCE = "reduce"


class DistributedInitMethod(StrEnum):
    TCP = "tcp"
    FILE = "file"


@dataclass
class SlurmConfig:
    mem_gb: int = 80
    timeout_hr: int = 168
    cpus_per_task: int = 8
    partition: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    qos: Optional[str] = None  # omegaconf in python 3.9 does not backport annotations
    account: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    additional_parameters: Optional[dict] = None


@dataclass
class SchedulerConfig:
    mode: SchedulerType = SchedulerType.LOCAL
    distributed_init_method: DistributedInitMethod = DistributedInitMethod.TCP
    ranks_per_node: int = 1
    num_nodes: int = 1
    num_array_jobs: int = 1
    slurm: SlurmConfig = field(default_factory=lambda: SlurmConfig)


@dataclass
class SlurmEnv:
    # reflects the job_id given by submitit (slurm id with array job id and array task id if they exist)
    job_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # reflects SLURM_JOB_ID only
    raw_job_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # SLURM_ARRAY_JOB_ID
    array_job_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # SLURM_ARRAY_TASK_ID
    array_task_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # reflects SLURM_RESTART_COUNT env variable
    restart_count: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )


@dataclass
class Metadata:
    # read-only metadata about the job, not user inputs
    commit: str
    log_dir: str
    checkpoint_dir: str
    results_dir: str
    config_path: str
    preemption_checkpoint_dir: str
    cluster_name: str
    array_job_num: int = 0
    slurm_env: SlurmEnv = field(default_factory=lambda: SlurmEnv())


@dataclass
class JobConfig:
    run_name: str = field(
        default_factory=lambda: get_timestamp_uid() + uuid.uuid4().hex.upper()[0:4]
    )
    timestamp_id: str = field(default_factory=lambda: get_timestamp_uid())
    run_dir: str = field(default_factory=lambda: tempfile.TemporaryDirectory().name)
    device_type: DeviceType = DeviceType.CUDA
    debug: bool = False
    scheduler: SchedulerConfig = field(default_factory=lambda: SchedulerConfig)
    logger: Optional[dict] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    seed: int = 0
    deterministic: bool = False
    runner_state_path: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # read-only metadata about the job, not user inputs
    metadata: Optional[Metadata] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    graph_parallel_group_size: Optional[int] = None

    def __post_init__(self) -> None:
        self.run_dir = os.path.abspath(self.run_dir)
        try:
            cluster = clusterscope.cluster()
        except RuntimeError:
            cluster = ""
        self.metadata = Metadata(
            commit=get_commit_hash(),
            log_dir=os.path.join(self.run_dir, self.timestamp_id, LOG_DIR_NAME),
            checkpoint_dir=os.path.join(
                self.run_dir, self.timestamp_id, CHECKPOINT_DIR_NAME
            ),
            results_dir=os.path.join(self.run_dir, self.timestamp_id, RESULTS_DIR),
            config_path=os.path.join(self.run_dir, self.timestamp_id, CONFIG_FILE_NAME),
            preemption_checkpoint_dir=os.path.join(
                self.run_dir,
                self.timestamp_id,
                CHECKPOINT_DIR_NAME,
                PREEMPTION_STATE_DIR_NAME,
            ),
            cluster_name=cluster,
        )


def _set_seeds(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def _set_deterministic_mode() -> None:
    # this is required for full cuda deterministic mode
    logging.info("Setting deterministic mode!")
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.use_deterministic_algorithms(True)


def _get_slurm_env() -> SlurmEnv:
    slurm_job_env = SlurmJobEnvironment()
    try:
        slurm_env = SlurmEnv(
            job_id=slurm_job_env.job_id,
            raw_job_id=slurm_job_env.raw_job_id,
            array_job_id=slurm_job_env.array_job_id,
            array_task_id=slurm_job_env.array_task_id,
            restart_count=os.environ.get("SLURM_RESTART_COUNT"),
        )
    except KeyError:
        # slurm environment variables are undefined, running locally
        slurm_env = SlurmEnv()

    return slurm_env


def remove_runner_state_from_submission(log_folder: str, job_id: str) -> None:
    # (HACK) Decouple the job from the runner state by manually modifying it
    # this ensures the saved runner state is not re-submitted in the event of a node failure
    # ie: if the job was started at state t=T, a requeue during node failure would resubmit the job
    # starting at state t=T again without calling the checkpoint callback, losing all progress in between.
    job_path = JobPaths(folder=log_folder, job_id=job_id)
    if os.path.isfile(job_path.submitted_pickle):
        submission_obj = DelayedSubmission.load(job_path.submitted_pickle)
        submission_obj.args[0].job.runner_state_path = None
        cloudpickle_dump(submission_obj, job_path.submitted_pickle)


class Submitit(Checkpointable):
    def __init__(self) -> None:
        self.config = None
        self.runner = None
        self.reducer = None

    def __call__(
        self, dict_config: DictConfig, run_type: RunType = RunType.RUN
    ) -> None:
        self.config = dict_config
        self.run_type = run_type
        # modify the config metadata to add slurm info if they exist
        self.config.job.metadata.slurm_env = _get_slurm_env()

        setup_env_vars()
        setup_logging()

        dist_config = map_job_config_to_dist_config(self.config.job)
        logging.info("Setting up distributed backend...")
        distutils.setup(dist_config)
        distutils.synchronize()
        if (
            distutils.is_master()
            and self.config.job.scheduler.mode == SchedulerType.SLURM
        ):
            # this pickle file is shared across all processes so can only modify this on the main rank
            remove_runner_state_from_submission(
                dict_config.job.metadata.log_dir,
                self.config.job.metadata.slurm_env.job_id,
            )

        if self.config.job.graph_parallel_group_size is not None:
            logging.info("Setting up graph parallel...")
            gp_utils.setup_graph_parallel_groups(
                self.config.job.graph_parallel_group_size,
                dist_config["distributed_backend"],
            )

        self._init_logger()

        _set_seeds(self.config.job.seed)
        if self.config.job.deterministic:
            _set_deterministic_mode()

        if run_type == RunType.RUN:
            logging.info("Calling runner.run() ...")
            self.runner: Runner = hydra.utils.instantiate(self.config.runner)
            self.runner.job_config = self.config.job
            # must call resume state AFTER the runner has been initialized
            self.runner.load_state(self.config.job.runner_state_path)
            self.runner.run()
        elif run_type == RunType.REDUCE:
            logging.info("Calling reducer.reduce() ...")
            self.reducer: Reducer = hydra.utils.instantiate(self.config.reducer)
            self.reducer.job_config = self.config.job
            self.reducer.runner_config = self.config.runner
            # must call resume state AFTER the runner has been initialized
            self.reducer.load_state(self.config.job.runner_state_path)
            self.reducer.reduce()
        else:
            raise ValueError(f"run type {run_type} is not recognized!")

        distutils.cleanup()

    def _init_logger(self) -> None:
        if (
            self.config.job.logger
            and distutils.is_master()
            and not self.config.job.debug
            and self.config.job.metadata.array_job_num == 0
        ):
            # get a partial function from the config and instantiate wandb with it
            # currently code assumes that we only use the WandBSingletonLogger
            logger_initializer = hydra.utils.instantiate(self.config.job.logger)
            simple_config = OmegaConf.to_container(
                self.config, resolve=True, throw_on_missing=True
            )
            logger_initializer(
                config=simple_config,
                run_id=self.config.job.timestamp_id,
                run_name=self.config.job.run_name,
                log_dir=self.config.job.metadata.log_dir,
            )

    def checkpoint(self, *args, **kwargs) -> DelayedSubmission:
        logging.error("Submitit checkpointing callback is triggered")
        save_path = self.config.job.metadata.preemption_checkpoint_dir
        cfg_copy = self.config.copy()
        # only assign if the save was successful
        cfg_copy.job.runner_state_path = None

        if (
            self.run_type == RunType.RUN
            and self.runner.save_state(save_path, is_preemption=True)
        ) or (
            self.run_type == RunType.REDUCE
            and self.reducer.save_state(save_path, is_preemption=True)
        ):
            cfg_copy.job.runner_state_path = save_path

        if WandBSingletonLogger.initialized():
            WandBSingletonLogger.get_instance().mark_preempting()
        logging.info(
            f"Submitit checkpointing callback is completed, resuming with use the following state: {save_path}"
        )
        return DelayedSubmission(Submitit(), cfg_copy)


def map_job_config_to_dist_config(job_cfg: JobConfig) -> dict:
    scheduler_config = job_cfg.scheduler
    return {
        "world_size": scheduler_config.num_nodes * scheduler_config.ranks_per_node,
        "distributed_backend": (
            "gloo" if job_cfg.device_type == DeviceType.CPU else "nccl"
        ),
        "submit": scheduler_config.mode == SchedulerType.SLURM,
        "cpu": job_cfg.device_type == DeviceType.CPU,
        "init_method": scheduler_config.distributed_init_method,
        # for distributed shared file initialization
        "shared_file_dir": os.path.join(job_cfg.run_dir, job_cfg.timestamp_id),
        "array_job_num": job_cfg.metadata.array_job_num,
    }


def get_canonical_config(config: DictConfig) -> DictConfig:
    # manually initialize metadata, because OmegaConf currently doesn't call __post_init__ on dataclasses
    job = OmegaConf.to_object(config.job)
    job.__post_init__()
    config.job = job
    # check that each key other than the allowed top level keys are used in config
    # find all top level keys are not in the allowed set
    all_keys = set(config.keys()).difference(ALLOWED_TOP_LEVEL_KEYS)
    used_keys = set()
    for key in all_keys:
        # make a copy of all keys except the key in question
        copy_cfg = OmegaConf.create({k: v for k, v in config.items() if k != key})
        try:
            OmegaConf.resolve(copy_cfg)
        except InterpolationKeyError:
            # if this error is thrown, this means the key was actually required
            used_keys.add(key)

    unused_keys = all_keys.difference(used_keys)
    if unused_keys != set():
        raise ValueError(
            f"Found unused keys in the config: {unused_keys}, please remove them!, only keys other than {ALLOWED_TOP_LEVEL_KEYS} or ones that are used as variables are allowed."
        )

    # resolve the config to fully replace the variables and delete all top level keys except for the ALLOWED_TOP_LEVEL_KEYS
    OmegaConf.resolve(config)
    return OmegaConf.create(
        {k: v for k, v in config.items() if k in ALLOWED_TOP_LEVEL_KEYS}
    )


def get_hydra_config_from_yaml(
    config_yml: str, overrides_args: list[str]
) -> DictConfig:
    # Load the configuration from the file
    os.environ["HYDRA_FULL_ERROR"] = "1"
    config_directory = os.path.dirname(os.path.abspath(config_yml))
    config_name = os.path.basename(config_yml)
    hydra.initialize_config_dir(config_directory, version_base="1.1")
    cfg = hydra.compose(config_name=config_name, overrides=overrides_args)
    # merge default structured config with initialized job object
    cfg = OmegaConf.merge({"job": OmegaConf.structured(JobConfig)}, cfg)
    # canonicalize config (remove top level keys that just used replacing variables)
    return get_canonical_config(cfg)


def _runner_wrapper(config: DictConfig, run_type: RunType = RunType.RUN):
    # This is needed when using elastic_launch for local runs since it looks for
    # the __name__ attribute of the function, Submitit.__call__ does not have one
    Submitit()(config, run_type)


def main(
    args: argparse.Namespace | None = None, override_args: list[str] | None = None
):
    if args is None:
        parser = argparse.ArgumentParser()
        parser.add_argument("-c", "--config", type=str, required=True)
        args, override_args = parser.parse_known_args()

    cfg = get_hydra_config_from_yaml(args.config, override_args)
    log_dir = cfg.job.metadata.log_dir
    os.makedirs(cfg.job.run_dir, exist_ok=True)
    os.makedirs(log_dir, exist_ok=True)

    OmegaConf.save(cfg, cfg.job.metadata.config_path)
    logging.info(f"saved canonical config to {cfg.job.metadata.config_path}")

    scheduler_cfg = cfg.job.scheduler
    logging.info(f"Running fairchemv2 cli with {cfg}")
    if scheduler_cfg.mode == SchedulerType.SLURM:  # Run on cluster
        assert (
            os.getenv("SLURM_SUBMIT_HOST") is None
        ), "SLURM DID NOT SUBMIT JOB!! Please do not submit jobs from an active slurm job (srun or otherwise)"
        executor = AutoExecutor(folder=log_dir, slurm_max_num_timeout=3)
        executor.update_parameters(
            name=cfg.job.run_name,
            mem_gb=scheduler_cfg.slurm.mem_gb,
            timeout_min=scheduler_cfg.slurm.timeout_hr * 60,
            slurm_partition=scheduler_cfg.slurm.partition,
            gpus_per_node=scheduler_cfg.ranks_per_node,
            cpus_per_task=scheduler_cfg.slurm.cpus_per_task,
            tasks_per_node=scheduler_cfg.ranks_per_node,
            nodes=scheduler_cfg.num_nodes,
            slurm_qos=scheduler_cfg.slurm.qos,
            slurm_account=scheduler_cfg.slurm.account,
            slurm_additional_parameters=scheduler_cfg.slurm.additional_parameters,
        )
        if scheduler_cfg.num_array_jobs == 1:
            job = executor.submit(Submitit(), cfg)
            logging.info(
                f"Submitted job id: {cfg.job.timestamp_id}, slurm id: {job.job_id}, logs: {cfg.job.metadata.log_dir}"
            )
            jobs = [job]
        elif scheduler_cfg.num_array_jobs > 1:
            executor.update_parameters(
                slurm_array_parallelism=scheduler_cfg.num_array_jobs,
            )

            jobs = []
            with executor.batch():
                for job_number in range(scheduler_cfg.num_array_jobs):
                    _cfg = cfg.copy()
                    _cfg.job.metadata.array_job_num = job_number
                    job = executor.submit(Submitit(), _cfg)
                    jobs.append(job)
            logging.info(f"Submitted {len(jobs)} jobs: {jobs[0].job_id.split('_')[0]}")

        if "reducer" in cfg:
            job_id = jobs[0].job_id.split("_")[0]
            executor.update_parameters(
                name=f"{cfg.job.run_name}_reduce",
                # set a single node, or do we want the same config as the Runner or a separate JobConfig
                nodes=1,
                slurm_dependency=f"afterok:{job_id}",
                slurm_additional_parameters={
                    "kill-on-invalid-dep": "yes"
                },  # kill the reducer if run fails
            )
            executor.submit(Submitit(), cfg, RunType.REDUCE)
    else:
        from torch.distributed.launcher.api import LaunchConfig, elastic_launch

        if scheduler_cfg.num_nodes > 1:
            cfg.job.scheduler.num_nodes = 1
            logging.warning(
                f"You cannot use more than one node (scheduler_cfg.num_nodes={scheduler_cfg.num_nodes}) in LOCAL mode, over-riding to 1 node"
            )
        if scheduler_cfg.ranks_per_node > 1:
            logging.info(
                f"Running in local mode with {scheduler_cfg.ranks_per_node} ranks using device_type:{cfg.job.device_type}"
            )
            launch_config = LaunchConfig(
                min_nodes=1,
                max_nodes=1,
                nproc_per_node=scheduler_cfg.ranks_per_node,
                rdzv_backend="c10d",
                max_restarts=0,
            )
            elastic_launch(launch_config, _runner_wrapper)(cfg)
            if "reducer" in cfg:
                elastic_launch(launch_config, _runner_wrapper)(cfg, RunType.REDUCE)
        else:
            logging.info("Running in local mode without elastic launch")
            distutils.setup_env_local()
            Submitit()(cfg)
            if "reducer" in cfg:
                Submitit()(cfg, RunType.REDUCE)



from __future__ import annotations

import logging
import timeit

from fairchem.core.calculate.pretrained_mlip import pretrained_checkpoint_path_from_name
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.datasets.common_structures import get_fcc_carbon_xtal
from fairchem.core.units.mlip_unit.api.inference import (
    InferenceSettings,
)
from fairchem.core.units.mlip_unit.predict import ParallelMLIPPredictUnit

logging.basicConfig(level=logging.INFO)


def get_qps(data, predictor, warmups: int = 10, timeiters: int = 100):
    def timefunc():
        predictor.predict(data)

    for _ in range(warmups):
        timefunc()

    result = timeit.timeit(timefunc, number=timeiters)
    qps = timeiters / result
    ns_per_day = qps * 24 * 3600 / 1e6
    return qps, ns_per_day


def main():
    ppunit = ParallelMLIPPredictUnit(
        inference_model_path=pretrained_checkpoint_path_from_name("uma-s-1p1"),
        device="cuda",
        inference_settings=InferenceSettings(
            tf32=True,
            merge_mole=True,
            wigner_cuda=False,
            compile=False,
            activation_checkpointing=False,
            internal_graph_gen_version=2,
            external_graph_gen=False,
        ),
        server_config={"workers": 8},
    )
    atoms = get_fcc_carbon_xtal(5000)
    # calc = FAIRChemCalculator(ppunit, task_name="omol")
    # atoms.calc = calc
    # print(atoms.get_potential_energy())
    # ppunit.cleanup()
    atomic_data = AtomicData.from_ase(atoms, task_name=["omat"])
    logging.info("Starting profile")
    qps, ns_per_day = get_qps(atomic_data, ppunit, warmups=10, timeiters=10)
    logging.info(f"QPS: {qps}, ns/day: {ns_per_day}")


if __name__ == "__main__":
    main()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Script to generate gifs from traj

Note:
This is just a quick way to generate gifs and visalizations from traj, there are many parameters and settings in the code that people can vary to make visualizations better. We have chosen these settings as this seem to work fine for most of our systems.

Requirements:

povray
ffmpeg
ase==3.21

"""

from __future__ import annotations

import argparse
import multiprocessing as mp
import os

import ase.io
import numpy as np
from ase.data import covalent_radii
from ase.io.pov import get_bondpairs


def pov_from_atoms(mp_args) -> None:
    atoms, idx, out_path = mp_args
    # how many extra repeats to generate on either side to look infinite
    extra_cells = 2
    # try and guess which atoms are adsorbates since the tags aren't correct after running in vasp
    # ideally this would be fixed by getting the right adsorbate atoms from the initial configurations
    atoms_organic = np.array([atom.symbol in {"C", "H", "O", "N"} for atom in atoms])
    # get the bare surface (note: this will not behave correctly for nitrides/hydrides/carbides/etc)
    atoms_surface = atoms[~atoms_organic].copy()
    # replicate the bare surface
    atoms_surface = atoms_surface.repeat((extra_cells * 2 + 1, extra_cells * 2 + 1, 1))
    # make an image of the adsorbate in the center of the slab
    atoms_adsorbate = atoms[atoms_organic]
    atoms_adsorbate.positions += extra_cells * (atoms.cell[0, :] + atoms.cell[1, :])
    # add the adsorbate to the replicated surface, then center the positions on the adsorbate
    num_surface_atoms = len(atoms_surface)
    atoms_surface += atoms_adsorbate
    atoms_surface.positions -= atoms_adsorbate.positions.mean(axis=0)
    # only include bonds for the adsorbate atoms
    bondpairs = get_bondpairs(atoms_surface)
    bondpairs = [
        bond
        for bond in bondpairs
        if bond[0] >= num_surface_atoms and bond[1] >= num_surface_atoms
    ]
    # write the image with povray
    bbox = (-6.4, -4, 6.4, 4)  # clip to a small region around the adsorbate
    os.chdir(f"{out_path}")
    renderer = ase.io.write(
        "snapshot_%04i.pov" % idx,
        atoms_surface,
        povray_settings={
            "celllinewidth": 0,
            "canvas_height": 300,
            "textures": ["intermediate"] * len(atoms_surface),
            "bondatoms": bondpairs,
        },
        bbox=bbox,
        rotation="-40x",
        radii=covalent_radii[atoms_surface.numbers],
    )
    renderer.render()
    print(f"image {idx} completed!")


def parallelize_generation(traj_path, out_path: str, n_procs) -> None:
    # make the covalent radii for O/C/N a little smaller to make bonds visible
    covalent_radii[6] = covalent_radii[6] * 0.7
    covalent_radii[7] = covalent_radii[7] * 0.7
    covalent_radii[8] = covalent_radii[8] * 0.7

    # name of the folder containing images and gif
    file_name = os.path.basename(traj_path).split(".")[0]
    out_path = os.path.join(out_path, file_name)
    out_path = os.path.abspath(out_path)
    os.makedirs(out_path, exist_ok=True)

    atoms_list = ase.io.read(traj_path, ":")

    # parallelizing image generation
    mp_args_list = [(atoms, idx, out_path) for idx, atoms in enumerate(atoms_list)]
    pool = mp.Pool(processes=n_procs)
    pool.map(pov_from_atoms, mp_args_list)

    # creating gif
    os.system(
        f"ffmpeg -pattern_type glob -i '{out_path}/*.png' {out_path}/{file_name}.gif"
    )


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--traj-path", required=True, help="Path to traj file")
    parser.add_argument(
        "--out-path",
        required=True,
        help="Directory to save generated images and gif",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=1,
        help="Number of processes to be used",
    )
    return parser


if __name__ == "__main__":
    parser: argparse.ArgumentParser = get_parser()
    args: argparse.Namespace = parser.parse_args()

    parallelize_generation(args.traj_path, args.out_path, args.num_workers)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import gzip
import json
import os
import pickle
from typing import TYPE_CHECKING, Any, Callable

import numpy as np

from fairchem.core.components.calculate import CalculateRunner

if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator


class OMolRunner(CalculateRunner):
    """
    Runner for OMol's evaluation tasks.
    """

    def __init__(
        self,
        calculator: Calculator,
        input_data: dict,
        benchmark_name: str,
        benchmark: Callable,
    ):
        """
        Initialize the OMolRunner.

        Args:
            calculator: ASE calculator to use for energy and force calculations
            input_data: Input dictionary data
            benchmark_name: Name of the benchmark task
            benchmark: Benchmark function to evaluate the input data
        """
        with open(input_data, "rb") as f:
            input_data = pickle.load(f)
        self.result_glob_pattern = f"{benchmark_name}_*-*.json.gz"
        self.benchmark_name = benchmark_name
        self.benchmark = benchmark
        super().__init__(calculator=calculator, input_data=input_data)

        self.input_keys = list(input_data.keys())

    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        """
        Perform calculations on a subset of structures.

        Splits the input data into chunks and processes the chunk corresponding to job_num.

        Args:
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.

        Returns:
            list[dict[str, Any]] - List of dictionaries containing calculation results
        """
        chunk_indices = np.array_split(self.input_keys, num_jobs)[job_num]
        chunk_data = {x: self.input_data[x] for x in chunk_indices}
        all_results = self.benchmark(chunk_data, self.calculator)
        return all_results

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        """
        Write calculation results to a compressed JSON file.

        Args:
            results: List of dictionaries containing elastic properties
            results_dir: Directory path where results will be saved
            job_num: Index of the current job
            num_jobs: Total number of jobs
        """
        with gzip.open(
            os.path.join(
                results_dir, f"{self.benchmark_name}_{num_jobs}-{job_num}.json.gz"
            ),
            "wb",
        ) as f:
            f.write(json.dumps(results).encode("utf-8"))

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True
